# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterable

from typing_extensions import Self

from ... import utils as ut
from ...localize import _
from ...utils import EPError
from ..rawtrigger.consttype import ConstType
from .structarr import _EUDStructMetaclass
from .vararray import _InternalVArray


class EUDStruct(ut.ExprProxy, metaclass=_EUDStructMetaclass):
    __slots__ = ("_initialized",)
    _fields_: Iterable[str | tuple]

    def __init__(self, *args, _from=None, _static_initval=None, **kwargs) -> None:
        super().__setattr__("_initialized", False)
        fieldcount = len(type(self)._fielddict)

        if _from is not None:
            super().__init__(_InternalVArray(fieldcount).cast(_from))
            self._initialized = True
        else:
            if _static_initval is None:  # hacky way...
                _static_initval = [0] * fieldcount
            super().__init__(_InternalVArray(fieldcount)(_static_initval))
            self._initialized = True
            self.constructor_static(*args, **kwargs)

    # Helper function for alloc & free
    # Due to cyclic dependency we import objpool inside methods
    @classmethod
    def alloc(cls, *args, **kwargs):
        from ...collections.objpool import get_global_pool

        return get_global_pool().alloc(cls, *args, **kwargs)

    @classmethod
    def free(cls, data):
        from ...collections.objpool import get_global_pool

        return get_global_pool().free(cls, data)

    # Constructor & Destructor of classes
    def constructor(self):
        """Constructor for individual structures.

        Default constructor accepts no arguments, but derived classes may
        accept additional arguments.

        This function is called when
            - Argument is allocated from pool   (self.isPooled = True)
            - Argument is generated             (self.isPooled = False)

        You may choose to either allocate member from pool or just allocate
        members statically via self.isPooled.
        """
        pass

    def constructor_static(self, *args, **kwargs):
        """Specialized constructor for static variables.

        Static variable may not require allocation for member variables.
        Function may specialize their behavior by overriding this function"""
        self.constructor(*args, **kwargs)

    def destructor(self):
        """Destructor for individual structures.

        Destructor accepts no arguments. Destructor is called when
            - Manually called. (Ex: stack variable)
            - free() is called for object
        """
        pass

    @classmethod
    def cast(cls, _from):
        try:
            return cls(_from=_from)
        except TypeError:
            obj = cls.__new__(cls)
            EUDStruct.__init__(obj, _from=_from)
            return obj

    def Assign(self, other) -> Self:  # noqa: N802
        from ..variable import EUDVariable, SetVariables

        if not isinstance(self._value, EUDVariable):
            raise EPError(_("Can't assign {} to constant expression").format(other))
        if isinstance(other, type(self)):
            SetVariables([self._value, self._epd], [other, other._epd])
        elif isinstance(other, int) and other == 0:
            SetVariables([self._value, self._epd], [0, ut.EPD(0)])
        else:
            raise EPError(_("Can't assign {} to {}").format(other, self))
        return self

    # Initializer

    def copy(self):
        """Create struct clone"""
        basetype = type(self)
        inst = basetype()
        self.copyto(inst)
        return inst

    def setall(self, values):
        self._value.fill(values, assert_expected_values_len=len(self._fielddict))

    def copyto(self, inst):
        """Copy struct to other instance"""
        for i, _nametype in enumerate(self._fielddict):
            inst.set(i, self.get(i))

    # Field setter & getter

    def getfield(self, name):
        attrid, attrtype = self._fielddict[name]
        attr = self._value.get(attrid)
        if attrtype:
            return attrtype.cast(attr)
        else:
            return attr

    def setfield(self, name, value):
        attrid, attrtype = self._fielddict[name]
        if isinstance(attrtype, ConstType):
            value = attrtype.cast(value)
        self._value.set(attrid, value)

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            try:
                return self.getfield(name)
            except KeyError as e:
                raise AttributeError from e

    def __setattr__(self, name, value):
        if self._initialized:
            try:
                self.setfield(name, value)
            except KeyError:
                raise ut.EPError(_("Unknown field name {}").format(name))
        else:
            super().__setattr__(name, value)

    def __getitem__(self, name):
        raise TypeError(f"'{type(self)}' object is not subscriptable")

    def __setitem__(self, name, newvalue):
        raise TypeError(f"'{type(self)}' object does not support item assignment")

    def __iter__(self):
        raise TypeError(f"'{type(self)}' object is not iterable")

    # Utilities

    def __lshift__(self, rhs):
        raise ut.EPError(_("Cannot reassign another value to eudstruct."))

    # Specializations

    def iaddattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.iadditem(attrid, value)

    # FIXME: add operator for Subtract
    def isubtractattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.isubtractitem(attrid, value)

    def isubattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.isubitem(attrid, value)

    def imulattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.imulitem(attrid, value)

    def ifloordivattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ifloordivitem(attrid, value)

    def imodattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.imoditem(attrid, value)

    def ilshiftattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ilshiftitem(attrid, value)

    def irshiftattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.irshiftitem(attrid, value)

    def ipowattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ipowitem(attrid, value)

    def iandattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ianditem(attrid, value)

    def iorattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ioritem(attrid, value)

    def ixorattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.ixoritem(attrid, value)

    # FIXME: Add operator for x[i] = ~x[i]
    def iinvertattr(self, name, value):
        attrid, _ = self._fielddict[name]
        self.iinvertitem(attrid, value)

    # Attribute comparisons

    def eqattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.eqitem(attrid, value)

    def neattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.neitem(attrid, value)

    def leattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.leitem(attrid, value)

    def geattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.geitem(attrid, value)

    def ltattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.ltitem(attrid, value)

    def gtattr(self, name, value):
        attrid, _ = self._fielddict[name]
        return self.gtitem(attrid, value)



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import core as c
from .. import memio
from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.typehint import _Tech
from ..localize import _
from .offsetmap import (
    ByteMember,
    EPDOffsetMap,
    IconMember,
    RaceResearchMember,
    StatTextMember,
    WordMember,
)
from .player import TrgPlayer


class Tech(EPDOffsetMap, ConstType):
    __slots__ = ()
    mineralCost: ClassVar[WordMember] = WordMember("array", 0x656248)
    gasCost: ClassVar[WordMember] = WordMember("array", 0x6561F0)
    timeCost: ClassVar[WordMember] = WordMember("array", 0x6563D8)
    energyCost: ClassVar[WordMember] = WordMember("array", 0x656380)
    researchRequirementOffset: ClassVar[WordMember] = WordMember("array", 0x656198)
    techUseRequirementOffset: ClassVar[WordMember] = WordMember("array", 0x6562F8)
    icon: ClassVar[IconMember] = IconMember("array", 0x656430)
    label: ClassVar[StatTextMember] = StatTextMember("array", 0x6562A0)
    race: ClassVar[RaceResearchMember] = RaceResearchMember("array", 0x656488)
    researched: ClassVar[ByteMember] = ByteMember("array", 0x656350)  # UNUSED?
    broodWarFlag: ClassVar[ByteMember] = ByteMember("array", 0x6564B4)  # bool?

    @ut.classproperty
    def range(self):
        return (0, 43, 1)

    @classmethod
    def cast(cls, _from: _Tech):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Tech) -> None:
        super().__init__(c.EncodeTech(initval))

    def _current_tech_epd_subp(self, player: TrgPlayer | int):
        r: c.EUDVariable
        sc_researched = 0x58CF44
        bw_researched = 0x58F140
        t = self._value
        if isinstance(player, TrgPlayer):
            p = player._value
        else:
            p = player
        if isinstance(p, int) or isinstance(player, int):
            ut.ep_assert(
                0 <= p <= 11,
                _("To get current tech, player must be in range from P1 to P12"),
            )
            if isinstance(t, c.EUDVariable):
                # (case 1) player is const, tech is var
                q, r, update = self._get_stride_cache(1)  # type: ignore[assignment]
                update_start, update_restore, update_end = update
                branch, sc_branch, bw_branch = (c.Forward() for _ in range(3))
                epd, subp = c.EUDCreateVariables(2)
                # Trigger 1: update tech cache and initialize epd, subp
                c.RawTrigger(
                    nextptr=update_start,
                    actions=[
                        c.SetNextPtr(update_start, update_restore),
                        c.SetMemory(update_start + 348, c.SetTo, q.GetVTable()),
                        c.SetNextPtr(update_end, q.GetVTable()),
                        c.SetNextPtr(q.GetVTable(), r.GetVTable()),
                        c.SetNextPtr(r.GetVTable(), branch),
                        q.QueueAssignTo(epd),
                        r.QueueAssignTo(subp),
                    ],
                )
                # Trigger 2: conditional branch for SC and BW tech
                branch << c.RawTrigger(
                    nextptr=sc_branch,
                    conditions=t.AtLeast(24),
                    actions=c.SetNextPtr(branch, bw_branch),
                )
                # (SC) Trigger 3: add to epd
                sc_branch << c.NextTrigger()
                epd += p * 24 // 4
                end = c.Forward()
                c.SetNextTrigger(end)
                # (BW) Trigger 3: add to epd
                bw_branch << c.NextTrigger()
                epd += p * 20 // 4 - (24 // 4)
                end << c.NextTrigger()
            # (case 2) player is const, tech is const
            elif t < 24:
                quotient, remainder = divmod(sc_researched + t + p * 24, 4)
                return quotient + ut.EPD(0), remainder
            else:
                quotient, remainder = divmod(bw_researched + t - 24 + p * 20, 4)
                return quotient + ut.EPD(0), remainder

        elif isinstance(t, c.EUDVariable):
            # (case 3) player is var, tech is var
            q_tech, r_tech, update_tech = self._get_stride_cache(1)
            if not isinstance(r_tech, c.EUDVariable):
                assert False
            t_update_start, t_update_restore, t_update_end = update_tech
            q24, r24, update_player = player._get_stride_cache(24)
            q20, r20, _update = player._get_stride_cache(20)
            p_update_start, p_update_restore, p_update_end = update_player

            # Trigger 1: update tech cache
            update_p, tech_branch = c.Forward(), c.Forward()
            c.RawTrigger(
                nextptr=t_update_start,
                actions=[
                    c.SetNextPtr(t_update_start, t_update_restore),
                    c.SetMemory(t_update_start + 348, c.SetTo, update_p),
                    c.SetNextPtr(t_update_end, update_p),
                ],
            )
            # Trigger 2: update player cache and initialize epd, subp
            # Can't update both tech and player at the same time
            # because t_update_end and p_update_end can be aliased.
            sc_branch, bw_branch = c.Forward(), c.Forward()
            epd, subp = c.EUDCreateVariables(2)
            update_p << c.VProc(
                [q_tech, r_tech],
                [
                    q_tech.QueueAssignTo(epd),
                    r_tech.QueueAssignTo(subp),
                    c.SetNextPtr(p_update_start, p_update_restore),
                    c.SetMemory(p_update_start + 348, c.SetTo, tech_branch),
                    c.SetNextPtr(p_update_end, tech_branch),
                    # prepare conditional branch
                    c.SetNextPtr(tech_branch, sc_branch),
                ],
            )
            c.SetNextTrigger(p_update_start)
            # Trigger 3: conditional branch for SC and BW tech
            tech_branch << c.RawTrigger(
                nextptr=sc_branch,
                conditions=t.AtLeast(24),
                actions=c.SetNextPtr(tech_branch, bw_branch),
            )
            # (SC) Trigger 4: add to epd, subp
            sc_branch << c.NextTrigger()
            c.SetVariables(
                [epd, epd],
                [ut.EPD(sc_researched), q24],
                [c.Add, c.Add],
            )
            end = c.Forward()
            c.SetNextTrigger(end)
            # (BW) Trigger 4: add to epd, subp (subp can be negative)
            bw_branch << c.NextTrigger()
            c.SetVariables(
                [epd, epd],
                [ut.EPD(bw_researched) - 24 // 4, q20],
                [c.Add, c.Add],
            )
            end << c.NextTrigger()
        else:
            # (case 4) player is var, tech is const
            if t < 24:
                q, r, update = player._get_stride_cache(24)  # type: ignore[assignment]
                quotient, remainder = divmod(t, 4)
            else:
                q, r, update = player._get_stride_cache(20)  # type: ignore[assignment]
                quotient, remainder = divmod(t - 24, 4)
            update_start, update_restore, update_end = update
            nexttrg = c.Forward()
            c.RawTrigger(
                nextptr=update_start,
                actions=[
                    c.SetNextPtr(update_start, update_restore),
                    c.SetMemory(update_start + 348, c.SetTo, nexttrg),
                    c.SetNextPtr(update_end, nexttrg),
                ],
            )
            nexttrg << c.NextTrigger()
            if t < 24:
                epd = ut.EPD(sc_researched) + quotient + q
            else:
                epd = ut.EPD(bw_researched) + quotient + q
            subp = remainder

        return epd, subp

    def __getitem__(self, player: TrgPlayer | int) -> c.EUDVariable:
        from ..memio.bwepdio import _boolread_epd

        epd, subp = self._current_tech_epd_subp(player)
        return _boolread_epd(epd, subp)

    def __setitem__(self, player: TrgPlayer | int, level) -> None:
        epd, subp = self._current_tech_epd_subp(player)
        memio.f_bwrite_epd(epd, subp, level)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from abc import ABCMeta, abstractmethod
from typing import ClassVar, cast

from ... import core as c
from ... import utils as ut
from ...localize import _
from ...memio import muldiv4table


class EPDOffsetMap(ut.ExprProxy, metaclass=ABCMeta):
    __slots__ = ()
    _cast: ClassVar[bool] = False
    _has_array_member: ClassVar[bool] = False
    _update: ClassVar[
        dict[c.EUDVariable, tuple[c.Forward, c.Forward, c.Forward]]
    ] = {}
    _maxvalue: ClassVar[dict[c.EUDVariable, int]] = {}
    _stride_key: ClassVar[dict[c.EUDVariable, frozenset[int]]] = {}
    _stride_dict: ClassVar[
        dict[c.EUDVariable, dict[int, tuple[c.EUDVariable, c.EUDVariable | int]]]
    ] = {}
    _advance: ClassVar[dict[c.EUDVariable, tuple[c.Forward, c.Forward]]] = {}

    @ut.classproperty
    @abstractmethod
    def range(self) -> tuple[int, int, int]:
        ...

    @classmethod
    def cast(cls, _from):
        if type(_from) is cls:
            return _from
        EPDOffsetMap._cast = True
        return cls(_from)

    def __init__(self, epd: int | c.EUDVariable) -> None:
        if isinstance(epd, c.EUDVariable) and not EPDOffsetMap._cast:
            epd = c.EUDVariable() << epd
            # To ellide var copy of rvalue variable (rvalue optimization),
            # variable is not added in dict to not increment refcount.
            #
            # example)
            # const foo = function (): TrgUnit { return 0; };
            # var x = foo();  // this should not copy variable.
            #
            # if epd not in EPDOffsetMap._update:
            #     EPDOffsetMap._update[epd] = (c.Forward(), c.Forward(), c.Forward())
            #     EPDOffsetMap._maxvalue[epd] = 0
            #     EPDOffsetMap._stride_key[epd] = frozenset(())
            #     EPDOffsetMap._stride_dict[epd] = {}
        EPDOffsetMap._cast = False
        super().__init__(epd)

    def _get_stride_cache(
        self, stride: int
    ) -> tuple[
        c.EUDVariable, int | c.EUDVariable, tuple[c.Forward, c.Forward, c.Forward]
    ]:
        value = self._value
        # lazy calculate multiplication & division
        stride_dict, maxvalue = self._get_stride_dict_and_maxvalue()
        update_start, update_restore, update_end = EPDOffsetMap._update[value]
        if stride not in stride_dict or maxvalue < type(self).range[1]:
            q = c.EUDVariable()
            r: int | c.EUDVariable = 0 if stride % 4 == 0 else c.EUDVariable()
            stride_dict[stride] = (q, r)
            strides = frozenset(stride_dict.keys())
            EPDOffsetMap._stride_key[value] = strides
            bit_index = type(self).range[1].bit_length()
            maxvalue = (1 << bit_index) - 1
            EPDOffsetMap._maxvalue[value] = maxvalue

            derived_vars = []
            for stride in sorted(stride_dict.keys()):
                variables = stride_dict[stride]
                if stride % 4 == 0:
                    derived_vars.append(variables[0])
                else:
                    derived_vars.extend(variables)  # type: ignore[arg-type]

            update_start.Reset()
            update_restore.Reset()
            update_end.Reset()
            upd = muldiv4table._caller(bit_index - 1, strides)(value, *derived_vars)
            update_start << upd[0]
            update_restore << upd[1]
            varcount = muldiv4table.varcount_dict[strides]
            read_end = muldiv4table.muldiv_end_table[varcount]
            update_end << read_end

            c.PushTriggerScope()
            add_carry, add_end = None, None
            start_actions = [
                value.AddNumber(1),
                c.SetMemory(update_start + 16, c.Add, 1),
            ]
            for k in sorted(stride_dict.keys()):
                epd, subp = stride_dict[k]
                quotient, remainder = divmod(k, 4)
                if quotient:
                    start_actions.append(epd.AddNumber(quotient))
                if remainder:
                    if not isinstance(subp, c.EUDVariable):
                        assert False
                    start_actions.append(subp.AddNumber(remainder))
                    add_end = c.RawTrigger(
                        conditions=subp.AtLeast(4),
                        actions=[subp.SubtractNumber(4), epd.AddNumber(1)],
                    )
                    if add_carry is None:
                        add_carry = add_end
            c.PopTriggerScope()
            c.PushTriggerScope()
            if add_carry:
                add_start = c.RawTrigger(nextptr=add_carry, actions=start_actions)
            else:
                add_start = add_end = c.RawTrigger(nextptr=0, actions=start_actions)
            c.PopTriggerScope()

            advance_start, advance_end = EPDOffsetMap._advance[value]
            advance_start.Reset()
            advance_end.Reset()
            advance_start << add_start
            advance_end << add_end
        else:
            q, r = stride_dict[stride]

        return q, r, (update_start, update_restore, update_end)

    def _get_stride_dict_and_maxvalue(
        self,
    ) -> tuple[dict[int, tuple[c.EUDVariable, c.EUDVariable | int]], int]:
        value = self._value
        ut.ep_assert(isinstance(value, c.EUDVariable), "constant EPDOffsetMap")
        try:
            stride_dict = EPDOffsetMap._stride_dict[value]
        except KeyError:
            # See EPDOffsetMap.__init__
            stride_dict = {}
            maxvalue = (1 << type(self).range[1].bit_length()) - 1
            ut.ep_assert(maxvalue < 1024, _("Reference type can't have ArrayMember"))
            EPDOffsetMap._update[value] = (c.Forward(), c.Forward(), c.Forward())
            EPDOffsetMap._maxvalue[value] = maxvalue
            EPDOffsetMap._stride_key[value] = frozenset(())
            EPDOffsetMap._stride_dict[value] = stride_dict
            c.PushTriggerScope()
            advance_trg = c.RawTrigger(nextptr=0, actions=value.AddNumber(1))
            advance_start, advance_end = c.Forward(), c.Forward()
            advance_start << advance_end << advance_trg
            c.PopTriggerScope()
            EPDOffsetMap._advance[value] = (advance_start, advance_end)
        else:
            maxvalue = EPDOffsetMap._maxvalue[value]
        return stride_dict, maxvalue

    def _advance_trg(self) -> tuple[c.Forward, c.Forward]:
        value = self._value
        self._get_stride_dict_and_maxvalue()
        return EPDOffsetMap._advance[value]

    def __iadd__(self, other):
        if (
            type(self)._has_array_member
            and isinstance(self._value, c.EUDVariable)
            and isinstance(other, int)
            and other == 1
        ):
            add_start, add_end = self._advance_trg()
            nexttrg = c.Forward()
            c.RawTrigger(nextptr=add_start, actions=c.SetNextPtr(add_end, nexttrg))
            nexttrg << c.NextTrigger()
        else:
            super().__iadd__(other)
        return self

    def getepd(self, name: str) -> c.EUDVariable:
        from .memberkind import CSpriteKind, CUnitKind, IscriptKind, PositionKind

        member = type(self).__dict__[name]
        kind = member.kind
        ut.ep_assert(kind.size() == 4, _("Only dword can be read as epd"))
        epd = member._get_epd(self)[0]
        if kind in (CUnitKind, CSpriteKind):
            return kind.read_epd(epd, 0)
        if kind in (PositionKind, IscriptKind):
            raise ut.EPError(_("Only dword can be read as epd"))

        from ...memio import f_epdread_epd

        return f_epdread_epd(epd)

    def getdwepd(self, name: str) -> tuple[c.EUDVariable, c.EUDVariable]:
        from .memberkind import CSpriteKind, CUnitKind, IscriptKind, PositionKind

        member = type(self).__dict__[name]
        kind = member.kind
        ut.ep_assert(kind.size() == 4, _("Only dword can be read as epd"))
        epd = member._get_epd(self)[0]
        if kind in (CUnitKind, CSpriteKind):
            cinstance = kind.read_epd(epd, 0)
            return cast(c.EUDVariable, cinstance._ptr), cinstance
        if kind in (PositionKind, IscriptKind):
            raise ut.EPError(_("Only dword can be read as epd"))

        from ...memio import f_dwepdread_epd

        return f_dwepdread_epd(epd)

    def getpos(self, name: str) -> tuple[c.EUDVariable, c.EUDVariable]:
        from .memberkind import CSpriteKind, CUnitKind, IscriptKind

        member = type(self).__dict__[name]
        kind = member.kind
        ut.ep_assert(kind.size() == 4, _("Only dword can be read as position"))
        if kind in (CUnitKind, CSpriteKind, IscriptKind):
            raise ut.EPError(_("Only dword can be read as position"))

        from ...memio import f_posread_epd

        return f_posread_epd(member._get_epd(self)[0])

    def iaddattr(self, name: str, value) -> None:
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        member.kind.add_epd(epd, subp, member.kind.cast(value))

    # TODO: add operator for Subtract
    def isubtractattr(self, name: str, value) -> None:
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        member.kind.subtract_epd(epd, subp, member.kind.cast(value))

    def isubattr(self, name: str, value) -> None:
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        value = member.kind.cast(value)
        member.kind.add_epd(epd, subp, -member.kind.cast(value))

    def imulattr(self, name, value):
        raise AttributeError

    def ifloordivattr(self, name, value):
        raise AttributeError

    def imodattr(self, name, value):
        raise AttributeError

    def ilshiftattr(self, name, value):
        raise AttributeError

    def irshiftattr(self, name, value):
        raise AttributeError

    def ipowattr(self, name, value):
        raise AttributeError

    def iandattr(self, name, value):
        raise AttributeError

    def iorattr(self, name, value):
        raise AttributeError

    def ixorattr(self, name, value):
        raise AttributeError

    # FIXME: Add operator for x[i] = ~x[i]
    def iinvertattr(self, name, value):
        raise AttributeError

    # Attribute comparisons

    def eqattr(self, name: str, value):
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        mask = c.f_bitlshift((1 << (8 * member.kind.size())) - 1, 8 * subp)
        value = member.kind.cast(value)
        if not (isinstance(subp, int) and subp == 0):
            value = c.f_bitlshift(value, 8 * subp)
        if isinstance(subp, int) and isinstance(value, int) and value & ~mask:
            raise ut.EPError(
                _("{} is out of range({}) for {} Member {}").format(
                    value, mask, type(self), name
                )
            )
        return c.MemoryXEPD(epd, c.Exactly, value, mask)

    def neattr(self, name, value):
        raise AttributeError

    def leattr(self, name, value):
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        mask = c.f_bitlshift((1 << (8 * member.kind.size())) - 1, 8 * subp)
        value = member.kind.cast(value)
        if not (isinstance(subp, int) and subp == 0):
            value = c.f_bitlshift(value, 8 * subp)
        if isinstance(subp, int) and isinstance(value, int) and value & ~mask:
            raise ut.EPError(
                _("{} is out of range({}) for {} Member {}").format(
                    value, mask, type(self), name
                )
            )
        return c.MemoryXEPD(epd, c.AtMost, value, mask)

    def geattr(self, name, value):
        member = type(self).__dict__[name]
        epd, subp = member._get_epd(self)
        mask = c.f_bitlshift((1 << (8 * member.kind.size())) - 1, 8 * subp)
        value = member.kind.cast(value)
        if not (isinstance(subp, int) and subp == 0):
            value = c.f_bitlshift(value, 8 * subp)
        if isinstance(subp, int) and isinstance(value, int) and value & ~mask:
            raise ut.EPError(
                _("{} is out of range({}) for {} Member {}").format(
                    value, mask, type(self), name
                )
            )
        return c.MemoryXEPD(epd, c.AtLeast, value, mask)

    def ltattr(self, name, value):
        raise AttributeError

    def gtattr(self, name, value):
        raise AttributeError


c.PushTriggerScope()
_epdcache = c.NextTrigger()
_epdcache_ptr = c.EUDVariable()
_epdcache_epd = ut.EPD(_epdcache_ptr)
c.RawTrigger(
    nextptr=_epdcache_epd.GetVTable(),
    conditions=_epdcache_ptr.Exactly(0),
    actions=_epdcache_epd.SetNumber(0),
)
c.PopTriggerScope()

c.PushTriggerScope()
_ptrcache = c.NextTrigger()
_ptrcache_epd = c.EUDVariable()
_ptrcache_ptr = _ptrcache_epd * 4
_ptrcache_ptr += 0x58A364
c.RawTrigger(
    nextptr=_ptrcache_ptr.GetVTable(),
    conditions=_ptrcache_epd.Exactly(0),
    actions=_ptrcache_ptr.SetNumber(0),
)
c.PopTriggerScope()


def _epd_cache(ptr: c.EUDVariable, ret=None) -> c.EUDVariable:
    epd = ret if ret else c.EUDVariable(0)
    is_ptr_equal = ptr.Exactly(0)
    check, update, skip, end = (c.Forward() for _ in range(4))
    check << c.RawTrigger(
        nextptr=update, conditions=is_ptr_equal, actions=c.SetNextPtr(check, skip)
    )
    update << c.VProc(
        [ptr, _epdcache_ptr],
        [
            ptr.QueueAssignTo(_epdcache_ptr),
            _epdcache_ptr.SetDest(ut.EPD(is_ptr_equal) + 2),
            _epdcache_epd.SetDest(epd),
            c.SetNextPtr(_epdcache_epd.GetVTable(), end),
        ],
    )
    c.SetNextTrigger(_epdcache)
    skip << c.RawTrigger(actions=c.SetNextPtr(check, update))
    end << c.NextTrigger()
    return epd


def _ptr_cache(epd: c.EUDVariable, ret=None) -> c.EUDVariable:
    ptr = ret if ret else c.EUDVariable(0)
    is_epd_equal = epd.Exactly(0)
    check, update, skip, end = (c.Forward() for _ in range(4))
    check << c.RawTrigger(
        nextptr=update, conditions=is_epd_equal, actions=c.SetNextPtr(check, skip)
    )
    update << c.VProc(
        [epd, _ptrcache_epd],
        [
            epd.QueueAssignTo(_ptrcache_epd),
            _ptrcache_epd.SetDest(ut.EPD(is_epd_equal) + 2),
            _ptrcache_ptr.SetDest(ptr),
            c.SetNextPtr(_ptrcache_ptr.GetVTable(), end),
        ],
    )
    c.SetNextTrigger(_ptrcache)
    skip << c.RawTrigger(actions=c.SetNextPtr(check, update))
    end << c.NextTrigger()
    return ptr



import itertools

from helper import *

a_initials = ExprProxy([5] * 10)


@TestInstance
def test_varray():
    zero_sized_varray = EUDVArray(0)()

    a = EUDVArray(len(a_initials))(a_initials)
    for i in range(8):
        a[i] = 2**i

    test_equality(
        "VArray test1",
        [a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9]],
        [1, 2, 4, 8, 16, 32, 64, 128, 5, 5],
    )

    for i in EUDLoopRange(3, 6):
        a[i] = i * i * i

    test_equality(
        "VArray test2",
        [a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9]],
        [1, 2, 4, 27, 64, 125, 64, 128, 5, 5],
    )

    b = EUDVariable(a)
    c = EUDVArray(8).cast(b)

    test_equality(
        "VArray test3",
        [c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]],
        [1, 2, 4, 27, 64, 125, 64, 128],
    )

    v_sum1 = EUDVariable()
    v_sum1 << 0
    for i in EUDLoopRange(0, 8):
        v_sum1 += c[i]

    test_equality("VArray test4", [v_sum1], [415])

    v_sum2 = EUDVariable()
    v_sum2 << 0
    for i in EUDLoopRange(0, 8):
        v_sum2 += c[i] * i

    test_equality("VArray test5", [v_sum2, a[9]], [2252, 5])

    varrs = (EUDVArray(2)(), EUDVArray(2).cast(EUDVariable(EUDVArray(2)())))
    indices = (1, EUDVariable(1))
    values = (135, EUDVariable(135))
    for varr, index, value in itertools.product(varrs, indices, values):
        varr[index] = value
        testname = []
        testname.append("V" if IsEUDVariable(varr) else "C")
        testname.append("[")
        testname.append("V" if IsEUDVariable(index) else "C")
        testname.append("] = ")
        testname.append("V" if IsEUDVariable(value) else "C")
        test_equality("".join(testname), varr[index], 135)
        varr[index] = 0

    for varr in varrs:
        index = indices[1]
        varr[index] = index
        if IsEUDVariable(varr):
            testname = "V[x] = x aliased write"
        else:
            testname = "C[x] = x aliased write"
        test_equality("".join(testname), varr[index], 1)
        varr[index] = 0

    arr = EUDArray(4)
    index = EUDVariable()
    index << 0
    varr = EUDVArray(4)([4, 3, 5, 8])
    for v in varr:
        arr[index] = v
        index += 1

    test_equality(
        "iter(const EUDVArray)", [arr[0], arr[1], arr[2], arr[3]], [4, 3, 5, 8]
    )

    arr[0] = 0
    arr[1] = 0
    arr[2] = 0
    arr[3] = 0
    index << 0
    varr = EUDVArray(4).cast(EUDVariable(varr))
    for v in varr:
        arr[index] = v
        index += 1

    test_equality(
        "iter(var EUDVArray)", [arr[0], arr[1], arr[2], arr[3]], [4, 3, 5, 8]
    )



from helper import *

test_operator("Multiplication", lambda x, y: x * y)
test_operator("Division", lambda x, y: x // y)
test_operator("Remainder", lambda x, y: x % y)
test_operator("Bitwise and", lambda x, y: x & y)
test_operator("Bitwise or", lambda x, y: x | y)
test_operator("Bitwise xor", lambda x, y: x ^ y)
test_operator("lshift", lambda x, y: f_bitlshift(x, y % 32))
test_operator("rshift", lambda x, y: f_bitrshift(x, y % 32))
test_operator("lshift big rhs", lambda x, y: f_bitlshift(x, y))
test_operator("rshift big rhs", lambda x, y: f_bitrshift(x, y))


@TestInstance
def test_variable_inequal():
    a, b, c, d = EUDCreateVariables(4)
    a << 1
    b << 1
    c << 2
    d << 2

    test_assert("never", EUDNot([a != b]))
    test_assert("never2", EUDNot([(a - b).AtLeast(1)]))
    test_assert("never and never", EUDNot([a != b, c != d]))
    test_assert("never(c) and never(c)", EUDNot([Never(), Never()]))
    test_assert("never and never(c)", EUDNot([a != b, Never()]))
    test_assert("never(c) and never", EUDNot([Never(), c != d]))

    test_assert("__lt__", a < c)
    test_assert("__le__", a <= c)
    test_assert("__le__", a <= b)
    test_assert("__le__", EUDNot(c <= a))  # a < c



# TEST HELPER
import functools
import os as _os
import random as _random
import sys as _sys

_sys.path.insert(
    1, _os.path.abspath(_os.path.join(_os.path.dirname(__file__), ".."))
)


from eudplib import *
from eudplib.collections.eudarray import _use_ptr_array
from eudplib.utils.etc import _allow_epd_on_epd
from eudplib.eudlib.utilf.datadumper import _add_datadumper

_allow_epd_on_epd(False)
_testFailed = EUDLightVariable()
_testNum = EUDVariable()
_failedNum = EUDVariable()


@functools.cache
def _failedTest():
    return EUDArray(1024)


###############################################################
# Unit testing helpers
###############################################################


origcp = EUDVariable()


def setcp1():
    origcp << f_getcurpl()
    f_setcurpl(Player1)


def resetcp():
    f_setcurpl(origcp)


def test_assert(testname, condition):
    global _failedNum, _testNum

    setcp1()

    if EUDIf()(condition):
        f_simpleprint("\x07 - [ OK ]\x04", testname)
        test_wait(0)
    if EUDElse()():
        f_simpleprint("\x08 - [FAIL]", testname)
        failedTestDb = Db(testname)
        _failedTest()[_failedNum] = failedTestDb
        _testFailed << 1
        test_wait(24)
    EUDEndIf()

    resetcp()

    _testNum += 1
    test_wait(0)


def test_equality(testname, real, expt):
    global _failedNum, _testNum

    real = Assignable2List(real)
    expt = Assignable2List(expt)
    assert len(real) == len(expt)

    setcp1()
    if EUDIf()([r == e for r, e in zip(real, expt)]):
        f_simpleprint("\x07 - [ OK ]\x04", testname)
        test_wait(0)
    if EUDElse()():
        f_simpleprint("\x08 - [FAIL]", testname)
        f_simpleprint(" \x03   - \x04 Output   : ", *real)
        f_simpleprint(" \x03   - \x04 Expected : ", *expt)
        failedTestDb = Db(testname)
        _failedTest()[_failedNum] = failedTestDb
        _failedNum += 1
        test_wait(24)
    EUDEndIf()

    resetcp()

    f_setcurpl(origcp)

    _testNum += 1


def test_operator(testname, realf, exptf=None):
    if exptf is None:
        exptf = realf

    try:
        f = realf._bodyfunc
    except AttributeError:
        f = realf
    argcount = f.__code__.co_argcount

    @TestInstance
    def test_operator():
        inputs = [EUDVariable() for _ in range(argcount)]
        expt, real = [], []

        for i in range(20):
            rnums = [_random.randint(0, 0xFFFFFFFF) for _ in range(argcount)]
            SetVariables(inputs, rnums)
            real.append(realf(*inputs))
            expt.append(exptf(*rnums) & 0xFFFFFFFF)

        test_assert(
            "Operator test : %s" % testname,
            [r == e for r, e in zip(real, expt)],
        )


class expect_error:
    def __init__(self, error):
        self._error = error

    def __enter__(self):
        PushTriggerScope()

    def __exit__(self, type, e, traceback):
        PopTriggerScope()
        if isinstance(e, self._error):
            print(f" - {self._error.__name__} as expected: {e}")
            return True
        else:
            raise RuntimeError(f"{self._error.__name__} not thrown")


class expect_eperror(expect_error):
    def __init__(self):
        super().__init__(EPError)


###############################################################
# Performance testing helper
###############################################################


perf_basecount = 200000


def test_perf(testname, func, count):
    starttm = f_dwread_epd(EPD(0x51CE8C))

    if EUDLoopN()(count):
        func()
    EUDEndLoopN()
    test_wait(0)

    endtm = f_dwread_epd(EPD(0x51CE8C))

    elapsedTime = starttm - endtm
    averageTime = elapsedTime // count
    setcp1()
    f_simpleprint(
        "\x03" * 150 + "[PERF] \x04%s \x03* %d    \x05" % (testname, count),
        averageTime,
        "/",
        elapsedTime,
        spaced=False,
    )
    resetcp()
    test_wait(12)


###############################################################


def test_complete():
    setcp1()
    f_simpleprint("\x03" + "=" * 40)
    succNum = _testNum - _failedNum
    f_simpleprint("\x04  Test result : ", succNum, "/", _testNum, spaced=False)
    resetcp()


_testList = []


def TestInstance(func):
    print(" - Adding test instance %s" % func.__name__)
    _testList.append((func, func.__name__))
    return func


@EUDFunc
def _testmain():
    for testfunc, testname in _testList:
        _testFailed << 0
        f_simpleprint("\x03[TEST] 테스트를 진행합니다:", testname)
        testfunc()
        test_equality(f"{testname} 종료", f_dwread_epd(0), 0)
        f_dwwrite_epd(0, 0)
        Trigger(_testFailed == 1, _failedNum.AddNumber(1))

    test_complete()


def test_runall(testname):
    LoadMap("outputmap/basemap/basemap_strx.scx")
    SaveMap("outputmap/test_%s.scx" % testname, _testmain, sector_size=15)
    print(f"eudplib version {eudplibVersion()}")


def test_wait(time):
    DoActions(SetMemory(0x6509A0, SetTo, time))
    EUDDoEvents()



from ..collections.eudarray import EUDArray
from ..core import EUDFunc
from ..ctrlstru import (
    DoActions,
    EUDBreakIf,
    EUDElse,
    EUDEndIf,
    EUDEndInfLoop,
    EUDIf,
    EUDInfLoop,
)
from ..memio.rwcommon import br1, bw1
from .cp949_table import cp949_table

"""
KSC5601 -> Unicode 2.0 mapping table, compressed for the 94*94 codeset.
Generated based on  KSC5601.txt at
  ftp://ftp.unicode.org/Public/MAPPINGS/EASTASIA/KSC

Unlike kuten-table, needed offset is 33 (0x21) instead of
32 for 7-bit portion of each byte.  i.e., a Unicode
codepoint for KSC's codepoint (n, m) would be found at
index (n-33)*94+m-33.
"""


@EUDFunc
def f_cp949_to_utf8_cpy(dst, src):
    # Create conversion table
    cvtb = [0] * 65536
    for (ch1, ch2), tab in cp949_table:
        cvtb[ch1 + ch2 * 256] = tab
    cvtb = EUDArray(cvtb)

    br1.seekoffset(src)
    bw1.seekoffset(dst)

    if EUDInfLoop()():
        b1 = br1.readbyte()
        EUDBreakIf(b1 == 0)
        if EUDIf()(b1 < 128):
            dst += 1
            bw1.writebyte(b1)
        if EUDElse()():
            b2 = br1.readbyte()
            EUDBreakIf(b2 == 0)
            code = cvtb[b2 * 256 + b1]
            if EUDIf()(code <= 0x07FF):
                # Encode as 2-byte
                b1 = code // (1 << 6)
                b2 = code & 0b111111
                DoActions(
                    dst.AddNumber(2),
                    b1.AddNumber(0b11000000),
                    b2.AddNumber(0b10000000),
                )
                bw1.writebyte(b1)
                bw1.writebyte(b2)
            if EUDElse()():
                # Encode as 3-byte
                b1 = code // (1 << 12)
                b2 = (code // (1 << 6)) & 0b111111
                b3 = code & 0b111111
                DoActions(
                    dst.AddNumber(3),
                    b1.AddNumber(0b11100000),
                    b2.AddNumber(0b10000000),
                    b3.AddNumber(0b10000000),
                )
                bw1.writebyte(b1)
                bw1.writebyte(b2)
                bw1.writebyte(b3)
            EUDEndIf()
        EUDEndIf()
    EUDEndInfLoop()
    bw1.writebyte(0)

    return dst



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterable, Sequence

from .. import core as c
from .. import utils as ut
from ..core import Condition, ConstExpr
from ..core.rawtrigger.rawtriggerdef import _Action
from .tpatcher import _Condition, patch_condition


def _branch_sub(
    conditions: Sequence[Condition],
    ontrue: ConstExpr,
    onfalse: ConstExpr,
    *,
    _actions: _Action | None = None,
) -> None:
    """
    Reduced version of EUDBranch with following restructions.
    - All fields of conditions/actions should be constant.
    - type(conditions) is list
    - len(conditions) <= 16
    """
    ut.ep_assert(len(conditions) <= 16, "conditions count > 16")

    brtrg = c.Forward()
    tjtrg = c.Forward()
    brtrg << c.RawTrigger(
        nextptr=onfalse, conditions=conditions, actions=c.SetNextPtr(brtrg, tjtrg)
    )
    actions: _Action = c.SetNextPtr(brtrg, onfalse)
    if _actions:
        _actions = ut.FlattenList(_actions)
        _actions.append(actions)
        actions = _actions
    tjtrg << c.RawTrigger(nextptr=ontrue, actions=actions)


def EUDBranch(  # noqa: N802
    conditions: _Condition | Iterable[_Condition | Iterable],
    ontrue: ConstExpr,
    onfalse: ConstExpr,
    *,
    _actions: _Action | None = None,
) -> None:
    """Branch by whether conditions is satisfied or not.

    :param conditions: Nested list of conditions.
    :param ontrue: When all conditions are true, this branch is taken.
    :param onfalse: When any of the conditions are false, this branch is taken.
    :param _actions: When all conditions are true, those actions are executed.
    """
    conditions = ut.FlattenList(conditions)
    conds = list(map(patch_condition, conditions))

    if len(conds) == 0:  # Just jump
        if _actions:
            c.RawTrigger(nextptr=ontrue, actions=_actions)
        else:
            c.SetNextTrigger(ontrue)
        return

    # Check all conditions
    for i in range(0, len(conds), 16):
        subontrue = c.Forward()
        subonfalse = onfalse

        if i + 16 < len(conds):
            _branch_sub(conds[i : i + 16], subontrue, subonfalse)
            subontrue << c.NextTrigger()
        else:
            _branch_sub(conds[i:], subontrue, subonfalse, _actions=_actions)
            subontrue << ontrue



from typing import Literal, TypeAlias

DefIscriptDict: dict[str, int] = {
    "Scourge": 0,
    "Unknown1": 1,
    "Scourge Death": 2,
    "Scourge Explosion": 3,
    "Broodling": 4,
    "Broodling Remnants": 5,
    "Infested Terran": 6,
    "Infested Terran Explosion": 7,
    "Guardian Cocoon": 8,
    "Defiler": 9,
    "Defiler Remnants": 10,
    "Drone": 11,
    "Drone Remnants": 12,
    "Egg": 13,
    "Egg Remnants": 14,
    "Guardian": 15,
    "Guardian Birth": 16,
    "Guardian Death": 17,
    "Hydralisk": 18,
    "Hydralisk Remnants": 19,
    "Infested Kerrigan": 20,
    "Larva": 21,
    "Larva Remnants": 22,
    "Mutalisk": 23,
    "Mutalisk Death": 24,
    "Overlord": 25,
    "Overlord Death": 26,
    "Queen": 27,
    "Queen Death": 28,
    "Ultralisk": 29,
    "Ultralisk Remnants": 30,
    "Zergling": 31,
    "Zergling Remnants": 32,
    "Zerg Air Death Explosion": 33,
    "Zerg Building Explosion": 34,
    "Unknown35": 35,
    "Zerg Birth": 36,
    "Egg Spawn": 37,
    "Cerebrate": 38,
    "Infested Command Center": 39,
    "Spawning Pool": 40,
    "Evolution Chamber": 41,
    "Creep Colony": 42,
    "Hatchery": 43,
    "Hive": 44,
    "Lair": 45,
    "Sunken Colony": 46,
    "Mature Chrysalis": 47,
    "Greater Spire": 48,
    "Defiler Mound": 49,
    "Queen's Nest": 50,
    "Nydus Canal": 51,
    "Overmind With Shell": 52,
    "Overmind Remnants": 53,
    "Overmind Without Shell": 54,
    "Ultralisk Cavern": 55,
    "Extractor": 56,
    "Hydralisk Den": 57,
    "Spire": 58,
    "Spore Colony": 59,
    "Infested Command Center Overlay": 60,
    "Zerg Construction (Small)": 61,
    "Zerg Construction (Medium)": 62,
    "Zerg Building Morph": 63,
    "Zerg Construction (Large)": 64,
    "Zerg Building Spawn": 65,
    "Battlecruiser": 66,
    "Civilian": 67,
    "Dropship": 68,
    "Firebat": 69,
    "Ghost": 70,
    "Ghost Remnants": 71,
    "Ghost Death": 72,
    "Nuke Beam": 73,
    "Nuke Target Dot": 74,
    "Goliath (Base)": 75,
    "Goliath (Turret)": 76,
    "Sarah Kerrigan": 77,
    "Marine": 78,
    "Marine Remnants": 79,
    "Marine Death": 80,
    "Scanner Sweep": 81,
    "Wraith": 82,
    "Wraith Afterburners": 83,
    "SCV": 84,
    "Unknown85": 85,
    "Vulture": 86,
    "Spider Mine": 87,
    "Science Vessel (Base)": 88,
    "Science Vessel (Turret)": 89,
    "Siege Tank (Tank) Base": 90,
    "Siege Tank (Tank) Turret": 91,
    "Siege Tank (Siege) Base": 92,
    "Siege Tank (Siege) Turret": 93,
    "Academy": 94,
    "Academy Overlay": 95,
    "Barracks": 96,
    "Armory": 97,
    "Armory Overlay": 98,
    "Comsat Station": 99,
    "Comsat Connector": 100,
    "Comsat Overlay": 101,
    "Command Center": 102,
    "Command Center Overlay": 103,
    "Crashed Battlecruiser": 104,
    "Supply Depot": 105,
    "Supply Depot Overlay": 106,
    "Control Tower": 107,
    "Control Tower Connector": 108,
    "Control Tower Overlay": 109,
    "Unknown110": 110,
    "Factory": 111,
    "Factory Overlay": 112,
    "Covert Ops": 113,
    "Covert Ops Connector": 114,
    "Covert Ops Overlay": 115,
    "Ion Cannon": 116,
    "Machine Shop": 117,
    "Machine Shop Connector": 118,
    "Missile Turret (Base)": 119,
    "Missile Turret (Turret)": 120,
    "Physics Lab": 121,
    "Physics Lab Connector": 122,
    "Bunker": 123,
    "Bunker Overlay": 124,
    "Refinery": 125,
    "Science Facility": 126,
    "Science Facility Overlay": 127,
    "Nuclear Silo": 128,
    "Nuclear Silo Connector": 129,
    "Nuclear Silo Overlay": 130,
    "Nuclear Missile": 131,
    "Unknown132": 132,
    "Nuke Explosion": 133,
    "Starport": 134,
    "Starport Overlay": 135,
    "Engineering Bay": 136,
    "Engineering Bay Overlay": 137,
    "Terran Construction (Large)": 138,
    "Terran Construction (Medium)": 139,
    "Terran Construction (Small)": 140,
    "Addon Construction": 141,
    "Explosion (Small)": 142,
    "Explosion (Medium)": 143,
    "Explosion (Large)": 144,
    "Building Rubble Header": 145,
    "Arbiter": 146,
    "Archon Energy": 147,
    "Archon Being": 148,
    "Archon Swirl": 149,
    "Unknown150": 150,
    "Carrier": 151,
    "Dark Templar (Hero)": 152,
    "Dragoon": 153,
    "Dragoon Remnants": 154,
    "Interceptor": 155,
    "Probe": 156,
    "Shuttle": 157,
    "High Templar": 158,
    "Reaver": 159,
    "Scarab": 160,
    "Scout": 161,
    "Scout Engines": 162,
    "Zealot": 163,
    "Zealot Death": 164,
    "Observer": 165,
    "Templar Archives": 166,
    "Assimilator": 167,
    "Observatory": 168,
    "Unknown169": 169,
    "Citadel Of Adun": 170,
    "Forge": 171,
    "Forge Overlay": 172,
    "Gateway": 173,
    "Cybernetics Core": 174,
    "Cybrnetics Core Overlay": 175,
    "Khaydarin Crystal Formation": 176,
    "Nexus": 177,
    "Nexus Overlay": 178,
    "Photon Cannon": 179,
    "Arbiter Tribunal": 180,
    "Pylon": 181,
    "Robotics Facility": 182,
    "Shield Battery": 183,
    "Shield Battery Overlay": 184,
    "Stargate": 185,
    "Stargate Overlay": 186,
    "Stasis Cell/Prison": 187,
    "Robotics Support Bay": 188,
    "Temple": 189,
    "Fleet Beacon": 190,
    "Warp Anchor": 191,
    "Warp Flash Header": 192,
    "Warp Texture": 193,
    "Unknown194": 194,
    "Unknown195": 195,
    "Unknown196": 196,
    "Unknown197": 197,
    "Ragnasaur": 198,
    "Rhynadon": 199,
    "Bengalaas": 200,
    "Vespene Geyser": 201,
    "Vespene Geyser2": 202,
    "Vespene Geyser Shadow": 203,
    "Mineral Field Type1": 204,
    "Mineral Field Type2": 205,
    "Mineral Field Type3": 206,
    "Unknown207": 207,
    "Zerg Beacon Overlay": 208,
    "Terran Beacon Overlay": 209,
    "Protoss Beacon Overlay": 210,
    "Zerg Beacon": 211,
    "Protoss Beacon": 212,
    "Terran Beacon": 213,
    "Unknown214": 214,
    "Powerups Shadow Header": 215,
    "Flag": 216,
    "Psi Emitter": 217,
    "Data Disk": 218,
    "Crystals Shadows": 219,
    "Young Chrysalis": 220,
    "Ore Chunk": 221,
    "Ore Chunk2": 222,
    "Gas Sac": 223,
    "Gas Sac2": 224,
    "Gas Orb": 225,
    "Gas Orb2": 226,
    "Gas Tank": 227,
    "Gas Tank2": 228,
    "Archon Overlay": 229,
    "Particle Beam Hit": 230,
    "Dual Photon Blaster Hit": 231,
    "Anti-Matter Missile": 232,
    "Pulse Cannon": 233,
    "Phase Disruptor": 234,
    "STA/STS Photon Cannon Overlay": 235,
    "Psionic Storm": 236,
    "Fusion Cutter Hit": 237,
    "Gauss Rifle Hit": 238,
    "Gemini Missiles": 239,
    "Longbolt Missile": 240,
    "C-10 Canister Rifle Hit": 241,
    "Fragmentation Grenade": 242,
    "ATA/ATS Laser Battery/Burst Lasers": 243,
    "Unknown244": 244,
    "Lockdown Hit": 245,
    "Arclite Shock Cannon Hit": 246,
    "Yamato Gun": 247,
    "Yamato Gun Trail": 248,
    "Lockdown/EMP Shockwave Missile": 249,
    "Siege Tank (Tank) Turret Overlay": 250,
    "Siege Tank (Siege) Turret Overlay": 251,
    "Science Vessel Overlay": 252,
    "Hallucination Hit": 253,
    "Unknown254": 254,
    "Unknown255": 255,
    "Needle Spines Hit": 256,
    "Venom (Unused)": 257,
    "Subterranean Tentacle": 258,
    "Venom Hit (Unused)": 259,
    "Acid Spore": 260,
    "Acid Spore Hit": 261,
    "Guardian Attack Overlay": 262,
    "Unknown263": 263,
    "Glave Wurm": 264,
    "Glave Wurm Hit": 265,
    "Seeker Spores": 266,
    "Queen Spell Holder": 267,
    "Psionic Shockwave Hit": 268,
    "Glave Wurm Trail": 269,
    "Seeker Spores Overlay": 270,
    "Acid Spray": 271,
    "Unknown272": 272,
    "Longbolt/Halo/Gemini Missiles Trail": 273,
    "Burowing Dust": 274,
    "Shadow Header": 275,
    "Shield Overlay": 276,
    "Unknown277": 277,
    "Double Explosion": 278,
    "Nuclear Missile Death": 279,
    "Spider Mine Explosion": 280,
    "Vespene Geyser Smokes": 281,
    "Unknown282": 282,
    "Fragmentation Grenade Hit": 283,
    "Grenade Shot Smoke": 284,
    "Phase Disruptor/Anti-Matter Missile Hit": 285,
    "Scarab/Anti-Matter Missile Overlay": 286,
    "Scarab Hit": 287,
    "Cursor Marker": 288,
    "Circle Marker": 289,
    "Carrier Engines": 290,
    "Engines/Glow Header": 291,
    "White Circle": 292,
    "Battlecruiser Attack Overlay": 293,
    "ATA/ATS Laser Battery/Burst Lasers Hit": 294,
    "Unknown295": 295,
    "Plague Cloud": 296,
    "Plague Overlay": 297,
    "Consume": 298,
    "Dark Swarm": 299,
    "Defensive Matrix Overlay": 300,
    "Defensive Matrix Hit": 301,
    "Ensnare": 302,
    "Ensnare Overlay": 303,
    "Irradiate": 304,
    "Recall Field": 305,
    "Stasis Field Overlay": 306,
    "Stasis Field Hit": 307,
    "Recharge Shields (Large)": 308,
    "Recharge Shields (Small)": 309,
    "High Templar Glow": 310,
    "Needle Spines Overlay": 311,
    "Flamethrower": 312,
    "Gemini Missiles Explosion": 313,
    "Yamato Gun Overlay": 314,
    "Yamato Gun Hit": 315,
    "Unknown316": 316,
    "Psionic Storm Part Variant1": 317,
    "Psionic Storm Part Variant2": 318,
    "EMP Shockwave Hit (Part1)": 319,
    "EMP Shockwave Hit (Part2)": 320,
    "Hallucination Death": 321,
    "Flames (Small)": 322,
    "Unknown323": 323,
    "Bleeding (Small) Variant1": 324,
    "Bleeding (Small) Variant2": 325,
    "Flames (Large)": 326,
    "Unknown327": 327,
    "Bleeding (Large) Variant1": 328,
    "Bleeding (Large) Variant2": 329,
    "Dust Variant1": 330,
    "Dust Variant2": 331,
    "Confirm Circle": 332,
    "Psi Field Type1": 333,
    "Psi Field Type2": 334,
    "Start Location": 335,
    "Doodad Header": 336,
    "Doodad Header2": 337,
    "Space Platform Doodad": 338,
    "Space Platform Doodad2": 339,
    "Installation Doodad": 340,
    "Installation Doodad2": 341,
    "Installation Right Wall Fans": 342,
    "Installation Left Wall Fans": 343,
    "Installation Gear": 344,
    "Floor Missile Trap": 345,
    "Floor Missile Trap Turret": 346,
    "Floor Gun Trap": 347,
    "Wall Missile Trap Type1": 348,
    "Wall Missile Trap Typet2": 349,
    "Wall Flame Trap Type1": 350,
    "Wall Flame Trap Type2": 351,
    "Map Revealer": 352,
    "Lurker Egg": 353,
    "Lurker": 354,
    "Unknown355": 355,
    "Lurker Remnants": 356,
    "Devourer": 357,
    "Devourer Birth": 358,
    "Devourer Death": 359,
    "Medic": 360,
    "Medic Remnants": 361,
    "Valkyrie": 362,
    "Unknown363": 363,
    "Unknown364": 364,
    "Dark Archon Energy": 365,
    "Dark Archon Being": 366,
    "Dark Archon Swirl": 367,
    "Dark Archon Death": 368,
    "Corsair": 369,
    "Corsair Engines": 370,
    "Unknown371": 371,
    "Dark Templar (Unit)": 372,
    "Neutron Flare": 373,
    "Disruption Web": 374,
    "Scantid": 375,
    "Kakaru": 376,
    "Ursadon": 377,
    "Halo Rocket": 378,
    "Optical Flare": 379,
    "Subterranean Spines": 380,
    "Corrosive Acid": 381,
    "Corrosive Acid Hit": 382,
    "Acid Spores (1) Overlay": 383,
    "Acid Spores (2-3) Overlay": 384,
    "Acid Spores (4-5) Overlay": 385,
    "Acid Spores (6-9) Overlay": 386,
    "Ice Doodad": 387,
    "Doodad Shadows Header (BW)": 388,
    "Restoration Hit": 389,
    "Mind Control Hit": 390,
    "Optical Flare Hit": 391,
    "Feedback": 392,
    "Maelstrom Overlay": 393,
    "Unknown394": 394,
    "Unknown395": 395,
    "Desert Doodad": 396,
    "Desert Doodad2": 397,
    "Desert Doodad3": 398,
    "Desert Doodad Overlay": 399,
    "Desert Doodad4": 400,
    "Twilight Doodad": 401,
    "Twilight Doodad2": 402,
    "Twilight Doodad3": 403,
    "Twilight Doodad4": 404,
    "Overmind Cocoon": 405,
    "Warp Gate": 406,
    "Psi Disrupter": 407,
    "Power Generator": 408,
    "Warp Gate Overlay": 409,
    "Xel'Naga Temple": 410,
    "Maelstrom Hit": 411,
}

DefaultIscript: TypeAlias = Literal[
    "Scourge",
    "Unknown1",
    "Scourge Death",
    "Scourge Explosion",
    "Broodling",
    "Broodling Remnants",
    "Infested Terran",
    "Infested Terran Explosion",
    "Guardian Cocoon",
    "Defiler",
    "Defiler Remnants",
    "Drone",
    "Drone Remnants",
    "Egg",
    "Egg Remnants",
    "Guardian",
    "Guardian Birth",
    "Guardian Death",
    "Hydralisk",
    "Hydralisk Remnants",
    "Infested Kerrigan",
    "Larva",
    "Larva Remnants",
    "Mutalisk",
    "Mutalisk Death",
    "Overlord",
    "Overlord Death",
    "Queen",
    "Queen Death",
    "Ultralisk",
    "Ultralisk Remnants",
    "Zergling",
    "Zergling Remnants",
    "Zerg Air Death Explosion",
    "Zerg Building Explosion",
    "Unknown35",
    "Zerg Birth",
    "Egg Spawn",
    "Cerebrate",
    "Infested Command Center",
    "Spawning Pool",
    "Evolution Chamber",
    "Creep Colony",
    "Hatchery",
    "Hive",
    "Lair",
    "Sunken Colony",
    "Mature Chrysalis",
    "Greater Spire",
    "Defiler Mound",
    "Queen's Nest",
    "Nydus Canal",
    "Overmind With Shell",
    "Overmind Remnants",
    "Overmind Without Shell",
    "Ultralisk Cavern",
    "Extractor",
    "Hydralisk Den",
    "Spire",
    "Spore Colony",
    "Infested Command Center Overlay",
    "Zerg Construction (Small)",
    "Zerg Construction (Medium)",
    "Zerg Building Morph",
    "Zerg Construction (Large)",
    "Zerg Building Spawn",
    "Battlecruiser",
    "Civilian",
    "Dropship",
    "Firebat",
    "Ghost",
    "Ghost Remnants",
    "Ghost Death",
    "Nuke Beam",
    "Nuke Target Dot",
    "Goliath (Base)",
    "Goliath (Turret)",
    "Sarah Kerrigan",
    "Marine",
    "Marine Remnants",
    "Marine Death",
    "Scanner Sweep",
    "Wraith",
    "Wraith Afterburners",
    "SCV",
    "Unknown85",
    "Vulture",
    "Spider Mine",
    "Science Vessel (Base)",
    "Science Vessel (Turret)",
    "Siege Tank (Tank) Base",
    "Siege Tank (Tank) Turret",
    "Siege Tank (Siege) Base",
    "Siege Tank (Siege) Turret",
    "Academy",
    "Academy Overlay",
    "Barracks",
    "Armory",
    "Armory Overlay",
    "Comsat Station",
    "Comsat Connector",
    "Comsat Overlay",
    "Command Center",
    "Command Center Overlay",
    "Crashed Battlecruiser",
    "Supply Depot",
    "Supply Depot Overlay",
    "Control Tower",
    "Control Tower Connector",
    "Control Tower Overlay",
    "Unknown110",
    "Factory",
    "Factory Overlay",
    "Covert Ops",
    "Covert Ops Connector",
    "Covert Ops Overlay",
    "Ion Cannon",
    "Machine Shop",
    "Machine Shop Connector",
    "Missile Turret (Base)",
    "Missile Turret (Turret)",
    "Physics Lab",
    "Physics Lab Connector",
    "Bunker",
    "Bunker Overlay",
    "Refinery",
    "Science Facility",
    "Science Facility Overlay",
    "Nuclear Silo",
    "Nuclear Silo Connector",
    "Nuclear Silo Overlay",
    "Nuclear Missile",
    "Unknown132",
    "Nuke Explosion",
    "Starport",
    "Starport Overlay",
    "Engineering Bay",
    "Engineering Bay Overlay",
    "Terran Construction (Large)",
    "Terran Construction (Medium)",
    "Terran Construction (Small)",
    "Addon Construction",
    "Explosion (Small)",
    "Explosion (Medium)",
    "Explosion (Large)",
    "Building Rubble Header",
    "Arbiter",
    "Archon Energy",
    "Archon Being",
    "Archon Swirl",
    "Unknown150",
    "Carrier",
    "Dark Templar (Hero)",
    "Dragoon",
    "Dragoon Remnants",
    "Interceptor",
    "Probe",
    "Shuttle",
    "High Templar",
    "Reaver",
    "Scarab",
    "Scout",
    "Scout Engines",
    "Zealot",
    "Zealot Death",
    "Observer",
    "Templar Archives",
    "Assimilator",
    "Observatory",
    "Unknown169",
    "Citadel Of Adun",
    "Forge",
    "Forge Overlay",
    "Gateway",
    "Cybernetics Core",
    "Cybrnetics Core Overlay",
    "Khaydarin Crystal Formation",
    "Nexus",
    "Nexus Overlay",
    "Photon Cannon",
    "Arbiter Tribunal",
    "Pylon",
    "Robotics Facility",
    "Shield Battery",
    "Shield Battery Overlay",
    "Stargate",
    "Stargate Overlay",
    "Stasis Cell/Prison",
    "Robotics Support Bay",
    "Temple",
    "Fleet Beacon",
    "Warp Anchor",
    "Warp Flash Header",
    "Warp Texture",
    "Unknown194",
    "Unknown195",
    "Unknown196",
    "Unknown197",
    "Ragnasaur",
    "Rhynadon",
    "Bengalaas",
    "Vespene Geyser",
    "Vespene Geyser2",
    "Vespene Geyser Shadow",
    "Mineral Field Type1",
    "Mineral Field Type2",
    "Mineral Field Type3",
    "Unknown207",
    "Zerg Beacon Overlay",
    "Terran Beacon Overlay",
    "Protoss Beacon Overlay",
    "Zerg Beacon",
    "Protoss Beacon",
    "Terran Beacon",
    "Unknown214",
    "Powerups Shadow Header",
    "Flag",
    "Psi Emitter",
    "Data Disk",
    "Crystals Shadows",
    "Young Chrysalis",
    "Ore Chunk",
    "Ore Chunk2",
    "Gas Sac",
    "Gas Sac2",
    "Gas Orb",
    "Gas Orb2",
    "Gas Tank",
    "Gas Tank2",
    "Archon Overlay",
    "Particle Beam Hit",
    "Dual Photon Blaster Hit",
    "Anti-Matter Missile",
    "Pulse Cannon",
    "Phase Disruptor",
    "STA/STS Photon Cannon Overlay",
    "Psionic Storm",
    "Fusion Cutter Hit",
    "Gauss Rifle Hit",
    "Gemini Missiles",
    "Longbolt Missile",
    "C-10 Canister Rifle Hit",
    "Fragmentation Grenade",
    "ATA/ATS Laser Battery/Burst Lasers",
    "Unknown244",
    "Lockdown Hit",
    "Arclite Shock Cannon Hit",
    "Yamato Gun",
    "Yamato Gun Trail",
    "Lockdown/EMP Shockwave Missile",
    "Siege Tank (Tank) Turret Overlay",
    "Siege Tank (Siege) Turret Overlay",
    "Science Vessel Overlay",
    "Hallucination Hit",
    "Unknown254",
    "Unknown255",
    "Needle Spines Hit",
    "Venom (Unused)",
    "Subterranean Tentacle",
    "Venom Hit (Unused)",
    "Acid Spore",
    "Acid Spore Hit",
    "Guardian Attack Overlay",
    "Unknown263",
    "Glave Wurm",
    "Glave Wurm Hit",
    "Seeker Spores",
    "Queen Spell Holder",
    "Psionic Shockwave Hit",
    "Glave Wurm Trail",
    "Seeker Spores Overlay",
    "Acid Spray",
    "Unknown272",
    "Longbolt/Halo/Gemini Missiles Trail",
    "Burowing Dust",
    "Shadow Header",
    "Shield Overlay",
    "Unknown277",
    "Double Explosion",
    "Nuclear Missile Death",
    "Spider Mine Explosion",
    "Vespene Geyser Smokes",
    "Unknown282",
    "Fragmentation Grenade Hit",
    "Grenade Shot Smoke",
    "Phase Disruptor/Anti-Matter Missile Hit",
    "Scarab/Anti-Matter Missile Overlay",
    "Scarab Hit",
    "Cursor Marker",
    "Circle Marker",
    "Carrier Engines",
    "Engines/Glow Header",
    "White Circle",
    "Battlecruiser Attack Overlay",
    "ATA/ATS Laser Battery/Burst Lasers Hit",
    "Unknown295",
    "Plague Cloud",
    "Plague Overlay",
    "Consume",
    "Dark Swarm",
    "Defensive Matrix Overlay",
    "Defensive Matrix Hit",
    "Ensnare",
    "Ensnare Overlay",
    "Irradiate",
    "Recall Field",
    "Stasis Field Overlay",
    "Stasis Field Hit",
    "Recharge Shields (Large)",
    "Recharge Shields (Small)",
    "High Templar Glow",
    "Needle Spines Overlay",
    "Flamethrower",
    "Gemini Missiles Explosion",
    "Yamato Gun Overlay",
    "Yamato Gun Hit",
    "Unknown316",
    "Psionic Storm Part Variant1",
    "Psionic Storm Part Variant2",
    "EMP Shockwave Hit (Part1)",
    "EMP Shockwave Hit (Part2)",
    "Hallucination Death",
    "Flames (Small)",
    "Unknown323",
    "Bleeding (Small) Variant1",
    "Bleeding (Small) Variant2",
    "Flames (Large)",
    "Unknown327",
    "Bleeding (Large) Variant1",
    "Bleeding (Large) Variant2",
    "Dust Variant1",
    "Dust Variant2",
    "Confirm Circle",
    "Psi Field Type1",
    "Psi Field Type2",
    "Start Location",
    "Doodad Header",
    "Doodad Header2",
    "Space Platform Doodad",
    "Space Platform Doodad2",
    "Installation Doodad",
    "Installation Doodad2",
    "Installation Right Wall Fans",
    "Installation Left Wall Fans",
    "Installation Gear",
    "Floor Missile Trap",
    "Floor Missile Trap Turret",
    "Floor Gun Trap",
    "Wall Missile Trap Type1",
    "Wall Missile Trap Typet2",
    "Wall Flame Trap Type1",
    "Wall Flame Trap Type2",
    "Map Revealer",
    "Lurker Egg",
    "Lurker",
    "Unknown355",
    "Lurker Remnants",
    "Devourer",
    "Devourer Birth",
    "Devourer Death",
    "Medic",
    "Medic Remnants",
    "Valkyrie",
    "Unknown363",
    "Unknown364",
    "Dark Archon Energy",
    "Dark Archon Being",
    "Dark Archon Swirl",
    "Dark Archon Death",
    "Corsair",
    "Corsair Engines",
    "Unknown371",
    "Dark Templar (Unit)",
    "Neutron Flare",
    "Disruption Web",
    "Scantid",
    "Kakaru",
    "Ursadon",
    "Halo Rocket",
    "Optical Flare",
    "Subterranean Spines",
    "Corrosive Acid",
    "Corrosive Acid Hit",
    "Acid Spores (1) Overlay",
    "Acid Spores (2-3) Overlay",
    "Acid Spores (4-5) Overlay",
    "Acid Spores (6-9) Overlay",
    "Ice Doodad",
    "Doodad Shadows Header (BW)",
    "Restoration Hit",
    "Mind Control Hit",
    "Optical Flare Hit",
    "Feedback",
    "Maelstrom Overlay",
    "Unknown394",
    "Unknown395",
    "Desert Doodad",
    "Desert Doodad2",
    "Desert Doodad3",
    "Desert Doodad Overlay",
    "Desert Doodad4",
    "Twilight Doodad",
    "Twilight Doodad2",
    "Twilight Doodad3",
    "Twilight Doodad4",
    "Overmind Cocoon",
    "Warp Gate",
    "Psi Disrupter",
    "Power Generator",
    "Warp Gate Overlay",
    "Xel'Naga Temple",
    "Maelstrom Hit",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import sys
import warnings

from ..localize import _


class EPError(Exception):
    pass


# fmt: off
_ERR = _("Must put Trigger into onPluginStart, beforeTriggerExec or afterTriggerExec")  # noqa: E501
# fmt: on


class TriggerScopeError(EPError):
    def __init__(self, msg=_ERR, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class EPWarning(Warning):
    pass


def ep_assert(statement: object, message: object = "Assertion failed") -> None:
    if not statement:
        raise EPError(message)


def ep_eprint(*args: object, **kwargs) -> None:
    print(*args, file=sys.stderr, **kwargs)


def ep_warn(message: str) -> None:
    warnings.warn(message, EPWarning)



# Source: https://github.com/PaddlePaddle/PaddleSOT/blob/7a2a7f4bdb998ccbec6a386ed69be54d996e923e/sot/opcode_translator/executor/pycode_generator.py#L149-L332

import sys
import types
from collections.abc import Callable, Iterator
from dis import Instruction, get_instructions
from typing import Any


def get_pycode_attributes() -> list[str]:
    """
    Returns a list of attribute names for PyCodeObject.
    NOTE(SigureMo): The order should consistent with signature specified in code_doc
    3.8: https://github.com/python/cpython/blob/3.8/Objects/codeobject.c#L416-L421
    3.10: https://github.com/python/cpython/blob/3.10/Objects/codeobject.c#L523-L543
    3.11: https://github.com/python/cpython/blob/3.11/Objects/codeobject.c#L1494-L1516

    Returns:
        list[str]: The attribute names for PyCodeObject.
    """
    pycode_attributes = [
        "co_argcount",
        "co_posonlyargcount",
        "co_kwonlyargcount",
        "co_nlocals",
        "co_stacksize",
        "co_flags",
        "co_code",
        "co_consts",
        "co_names",
        "co_varnames",
        "co_filename",
        "co_name",
    ]
    if sys.version_info >= (3, 11):
        pycode_attributes.append("co_qualname")
    pycode_attributes.append("co_firstlineno")
    if sys.version_info >= (3, 10):
        pycode_attributes.append("co_linetable")
    else:
        pycode_attributes.append("co_lnotab")
    if sys.version_info >= (3, 11):
        pycode_attributes.append("co_exceptiontable")
    pycode_attributes += [
        "co_freevars",
        "co_cellvars",
    ]
    return pycode_attributes


PYCODE_ATTRIBUTES = get_pycode_attributes()


def gen_code_options(code: types.CodeType) -> dict[str, Any]:
    """
    Generates a dictionary of code options for the given code object.

    Args:
        code (types.CodeType): The code object.

    Returns:
        dict[str, any]: The code options.
    """
    code_options = {}
    for k in PYCODE_ATTRIBUTES:
        val = getattr(code, k)
        if isinstance(val, tuple):
            val = list(val)
        code_options[k] = val

    return code_options


def gen_new_opcode(
    code_object: types.CodeType,
    code_options: dict[str, Any],
    keys: list[str],
    ep_lineno_map: Callable[[int], int],
) -> types.CodeType:
    """
    Generates a new code object with the given instructions, code options, and keys.

    Args:
        code_object (types.CodeType): The original CodeType for instructions.
        code_options (dict[str, any]): The code options for the new code object.
        keys (list[str]): The keys to specify the order of code options.
        ep_lineno_map (Callable[[int], int]): The function to convert line number.

    Returns:
        types.CodeType: The new code object.
    """
    ep_firstlineno = ep_lineno_map(code_options["co_firstlineno"])
    linetable = assemble(
        get_instructions(code_object),
        ep_firstlineno,
        ep_lineno_map,
    )
    if sys.version_info >= (3, 10):
        # Python deprecated co_lnotab in 3.10, use co_linetable instead
        # https://peps.python.org/pep-0626/
        code_options["co_linetable"] = linetable
    else:
        code_options["co_lnotab"] = linetable
    code_options["co_firstlineno"] = ep_firstlineno
    code_options["co_nlocals"] = len(code_options["co_varnames"])
    # TODO: should we generate or empty 3.11 exception table?
    # if sys.version_info >= (3, 11):
    #     code_options["co_exceptiontable"] = bytes([])
    for key, val in code_options.items():
        if isinstance(val, list):
            code_options[key] = tuple(val)
    # code_options is a dict, use keys to make sure the input order
    return types.CodeType(*[code_options[k] for k in keys])


def assemble(
    instructions: Iterator[Instruction],
    firstlineno: int,
    ep_lineno_map: Callable[[int], int],
) -> bytes:
    """
    Assembles a list of instructions into lnotab.

    Args:
        instructions (Iterator[Instruction]): The list of instructions.
        firstlineno (int): The starting line number.
        ep_lineno_map (Callable[[int], int]): The function to convert line number.

    Returns:
        bytes: The assembled lnotab.
    """
    code: list[int] = []
    linetable = []
    ep_firstlineno = firstlineno

    calc_linetable, update_cursor = create_linetable_calculator(ep_firstlineno)

    for instr in instructions:
        # set linetable, Python 3.11 need to set linetable for each instruction
        if instr.starts_line is not None or sys.version_info >= (3, 11):
            ep_lineno: int | None
            if instr.starts_line is not None:
                ep_lineno = ep_lineno_map(instr.starts_line)
            else:
                ep_lineno = None
            linetable.extend(calc_linetable(ep_lineno, len(code)))
            update_cursor(ep_lineno, len(code))

        # get bytecode
        arg = instr.arg or 0
        code.extend((instr.opcode, arg & 0xFF))
        # fill CACHE
        for _ in range(get_instruction_size(instr) // 2 - 1):
            code.extend((0, 0))

    if sys.version_info >= (3, 11):
        # End hook for Python 3.11
        linetable.extend(calc_linetable(None, len(code)))
    elif sys.version_info >= (3, 10):
        # End hook for Python 3.10
        linetable.extend(calc_linetable(0, len(code)))

    return bytes(linetable)


def to_byte(num):
    """
    Converts a negative number to an unsigned byte.

    Args:
        num (int): The number to convert.

    Returns:
        int: The converted unsigned byte.
    """
    if num < 0:
        num += 256
    return num


# Cache for some opcodes, it's for Python 3.11+
# https://github.com/python/cpython/blob/3.11/Include/internal/pycore_opcode.h#L41-L53
PYOPCODE_CACHE_SIZE = {
    "BINARY_SUBSCR": 4,
    "STORE_SUBSCR": 1,
    "UNPACK_SEQUENCE": 1,
    "STORE_ATTR": 4,
    "LOAD_ATTR": 4,
    "COMPARE_OP": 2,
    "LOAD_GLOBAL": 5,
    "BINARY_OP": 1,
    "LOAD_METHOD": 10,
    "PRECALL": 1,
    "CALL": 4,
}


def get_instruction_size(instr) -> int:
    cache_size = 0
    if sys.version_info >= (3, 11):
        cache_size = PYOPCODE_CACHE_SIZE.get(instr.opname, 0)
    return 2 * (cache_size + 1)


def create_linetable_calculator(firstlineno: int):
    """
    Creates a line table calculator function.

    Args:
        firstlineno (int): The starting line number.

    Returns:
        Callable: The line table calculator function.
    """
    cur_lineno = firstlineno
    cur_bytecode = 0
    line_offset = 0  # For Python 3.10

    def update_cursor(starts_line: int | None, code_length: int):
        nonlocal cur_lineno, cur_bytecode
        cur_bytecode = code_length
        if starts_line is not None:
            cur_lineno = starts_line

    def calc_lnotab(starts_line: int, code_length: int):
        """
        Calculates the lnotab for Python 3.8 and 3.9.
        https://github.com/python/cpython/blob/3.9/Objects/lnotab_notes.txt

        Args:
            starts_line (int): The line number where the instruction starts.
            code_length (int): The length of the code.

        Returns:
            list[int]: The lnotab.
        """
        nonlocal cur_lineno, cur_bytecode
        line_offset = starts_line - cur_lineno
        byte_offset = code_length - cur_bytecode
        result: list[int] = []

        while line_offset or byte_offset:
            line_offset_step = min(max(line_offset, -128), 127)
            byte_offset_step = min(max(byte_offset, 0), 255)
            result.extend((byte_offset_step, to_byte(line_offset_step)))
            line_offset -= line_offset_step
            byte_offset -= byte_offset_step
        return result

    def calc_linetable_py310(starts_line: int, code_length: int):
        """
        Calculates the linetable for Python 3.10.
        https://github.com/python/cpython/blob/3.10/Objects/lnotab_notes.txt

        Args:
            starts_line (int): The line number where the instruction starts.
            code_length (int): The length of the code.

        Returns:
            list[int]: The linetable.
        """
        nonlocal cur_lineno, cur_bytecode, line_offset
        byte_offset = code_length - cur_bytecode
        result: list[int] = []
        while line_offset or byte_offset:
            line_offset_step = min(max(line_offset, -127), 127)
            byte_offset_step = min(max(byte_offset, 0), 254)
            result.extend((byte_offset_step, to_byte(line_offset_step)))
            line_offset -= line_offset_step
            byte_offset -= byte_offset_step
        line_offset = starts_line - cur_lineno
        return result

    def _encode_varint(num: int):
        """
        Encode unsigned integer into variable-length format.
        """
        continue_flag = 0b01 << 6
        stop_flag = 0b00 << 6
        while num >= 0x40:
            yield (num & 0x3F) | continue_flag
            num >>= 6
        yield num | stop_flag

    def _encode_svarint(num: int):
        """
        Encode signed integer into variable-length format.
        """
        unsigned_value = (((-num) << 1) | 1) if num < 0 else (num << 1)
        yield from _encode_varint(unsigned_value)

    def _encode_bytecode_to_entries_py311(line_offset: int, byte_offset: int):
        if not byte_offset:
            return []
        if 0 < byte_offset <= 8:
            entry_head = 0b1_1101_000 | (byte_offset - 1)
            return [entry_head, *list(_encode_svarint(line_offset))]
        return [
            *_encode_bytecode_to_entries_py311(line_offset, 8),
            *_encode_bytecode_to_entries_py311(line_offset, byte_offset - 8),
        ]

    def calc_linetable_py311(starts_line: int | None, code_length: int):
        """
        Calculates the linetable for Python 3.11.
        https://github.com/python/cpython/blob/3.11/Objects/locations.md

        Args:
            starts_line (int): The line number where the instruction starts.
            code_length (int): The length of the code.

        Returns:
            list[int]: The linetable.
        """
        nonlocal cur_lineno, cur_bytecode
        line_offset = starts_line - cur_lineno if starts_line is not None else 0
        byte_offset = (code_length - cur_bytecode) // 2
        return _encode_bytecode_to_entries_py311(line_offset, byte_offset)

    if sys.version_info >= (3, 11):
        return calc_linetable_py311, update_cursor
    elif sys.version_info >= (3, 10):
        return calc_linetable_py310, update_cursor
    else:
        return calc_lnotab, update_cursor



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing_extensions import Self

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...memio import f_bread_epd, f_dwread_epd
from ...scdata import TrgPlayer

_userp: TrgPlayer = TrgPlayer.cast(c.EUDVariable())
_userp_fws: set[tuple] = set()


def _f_inituserplayerid():
    f_bread_epd(ut.EPD(0x512684), 0, ret=[ut.EPD(_userp.getValueAddr())])


def f_getuserplayerid() -> TrgPlayer:
    return _userp


def IsUserCP() -> c.Condition:  # noqa: N802
    """Condition: check if CurrentPlayer equals to user player id (local)"""
    fw = c.Forward()
    ret = c.Memory(0x6509B0, c.Exactly, fw)
    fw << _UserPFw(ret, 8)
    return ret


def _action_all(action: c.Action) -> tuple[c.Action, c.Action]:
    fw = c.Forward()
    ret = c.SetMemory(0x6509B0, c.SetTo, fw)
    fw << _UserPFw(ret, 20)
    return ret, action


def DisplayTextAll(text) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: DisplayText for everyone (including observers)"""
    return _action_all(c.DisplayText(text))


def PlayWAVAll(soundpath) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: PlayWAV for everyone (including observers)"""
    return _action_all(c.PlayWAV(soundpath))


def MinimapPingAll(location) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: MinimapPing for everyone (including observers)"""
    return _action_all(c.MinimapPing(location))


def CenterViewAll(location) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: CenterView for everyone (including observers)"""
    return _action_all(c.CenterView(location))


def SetMissionObjectivesAll(text) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: SetMissionObjectives for everyone (including observers)"""
    return _action_all(c.SetMissionObjectives(text))


def TalkingPortraitAll(unit, time) -> tuple[c.Action, c.Action]:  # noqa: N802
    """Action: TalkingPortrait for everyone (including observers)"""
    return _action_all(c.TalkingPortrait(unit, time))


# NOTE: should we add (Un)MuteUnitSpeechAll?


class _UserPFw(c.ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(self, dest, offset) -> None:
        super().__init__()
        self._initobj = (dest, offset)

    def Evaluate(self):  # noqa: N802
        _register_userp(*self._initobj)
        return c.toRlocInt(0)


def _rcpc_reset_userp():
    _userp_fws.clear()


c.RegisterCreatePayloadCallback(_rcpc_reset_userp)


def _register_userp(dest, offset):
    _userp_fws.add((dest, offset))


class UserPBuffer(c.EUDObject):
    def __init__(self):
        super().__init__()

    def GetDataSize(self):  # noqa: N802
        return (len(_userp_fws) + 1) * 4

    def WritePayload(self, pbuf):  # noqa: N802
        for dest, offset in _userp_fws:
            pbuf.WriteDword(ut.EPD(dest + offset))
        pbuf.WriteDword(0xFFFFFFFF)


def _f_initisusercp():
    """(internal) Initialize IsUserCP."""
    rb = UserPBuffer()
    ptr = c.EUDVariable(ut.EPD(rb))
    userp = f_getuserplayerid()
    write = c.SetDeaths(0, c.SetTo, 0, 0)
    c.VProc(userp, userp.QueueAssignTo(ut.EPD(write) + 5))

    if cs.EUDInfLoop()():
        v = f_dwread_epd(ptr)
        cs.EUDBreakIf(v == 0xFFFFFFFF)
        c.VProc(v, v.SetDest(ut.EPD(write) + 4))
        c.RawTrigger(actions=[write, ptr.AddNumber(1)])

    cs.EUDEndInfLoop()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...collections.eudarray import EUDArray
from ...memio import f_dwread_epd, f_repmovsd_epd

_patch_max = 8192

dws_top, ps_top = c.EUDVariable(), c.EUDVariable()


@functools.cache
def patchstack():
    return EUDArray(3 * _patch_max)


@functools.cache
def dwstack():
    return EUDArray(_patch_max)


def pushpatchstack(value: c.EUDVariable | int) -> None:
    global ps_top
    patchstack()[ps_top] = value
    ps_top += 1


def poppatchstack() -> c.EUDVariable:
    global ps_top
    ps_top -= 1
    return patchstack()[ps_top]


@c.EUDFunc
def f_dwpatch_epd(dstepd, value):
    global dws_top

    prev_value = f_dwread_epd(dstepd)
    if dwstack()._value._is_epd():
        dws_pos = dwstack() + dws_top
    else:
        dws_pos = ut.EPD(dwstack()) + dws_top
    c.VProc(
        [dstepd, value, dws_pos, prev_value],
        [
            dstepd.SetDest(ut.EPD(value.getDestAddr())),
            dws_pos.SetDest(ut.EPD(prev_value.getDestAddr())),
        ],
    )

    pushpatchstack(dstepd)
    pushpatchstack(dws_pos)
    pushpatchstack(1)
    dws_top += 1


@c.EUDFunc
def f_blockpatch_epd(dstepd, srcepd, dwn):
    """Patch 4*dwn bytes of memory at dstepd with memory of srcepd.

    .. note::
        After calling this function, contents at srcepd memory may change.
        Since new contents are required for :py:`f_unpatchall` to run, you
        shouldn't use the memory for any other means.
    """

    global dws_top

    # Push to stack
    pushpatchstack(dstepd)
    pushpatchstack(srcepd)
    pushpatchstack(dwn)
    copydwn = c.EUDVariable()
    endbranch, contpoint = c.Forward(), c.Forward()
    c.SeqCompute(
        [
            (dws_top, c.Add, 1),
            (copydwn, c.SetTo, 256),
            (ut.EPD(endbranch) + 1, c.SetTo, contpoint),
        ]
    )

    # Swap contents btw dstepd, srcepd
    tmpbuffer = c.Db(1024)

    if cs.EUDWhileNot()(dwn == 0):
        endbranch << c.RawTrigger(
            nextptr=0,
            conditions=dwn <= 255,
            actions=[
                c.SetNextPtr(endbranch, dwn.GetVTable()),
                dwn.SetDest(copydwn),
                c.SetNextPtr(dwn.GetVTable(), contpoint),
            ],
        )
        contpoint << c.RawTrigger(actions=dwn.SubtractNumber(256))

        f_repmovsd_epd(ut.EPD(tmpbuffer), dstepd, copydwn)
        f_repmovsd_epd(dstepd, srcepd, copydwn)
        f_repmovsd_epd(srcepd, ut.EPD(tmpbuffer), copydwn)
    cs.EUDEndWhile()


@c.EUDFunc
def f_unpatchall():
    global ps_top, dws_top
    if cs.EUDWhile()(ps_top >= 1):
        dws_top -= 1
        dwn = poppatchstack()
        unpatchsrcepd = poppatchstack()
        dstepd = poppatchstack()
        f_repmovsd_epd(dstepd, unpatchsrcepd, dwn)
    cs.EUDEndWhile()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import utils as ut
from ...core.allocator.pbuffer import Payload
from ...core.mapdata.chktok import CHK
from ...core.mapdata.stringmap import get_string_map, get_string_section_name
from ...localize import _
from ...trigtrg import trigtrg as tt

""" Stage 1 : works in TRIG section.
- Check if EUD is available. If not, defeat
- Initialize stage 2 & execute it
- Modifies TRIG rawtrigger's nextptr. Modification is fixed in stage 3.
"""
trglist: list[bytes] = []


def _Trigger(  # noqa: N802
    *,
    # FIXME: Cannot determine type of "AllPlayers"  [has-type]
    players: list[tt.Player] = [17],
    conditions: list[bytes] | bytes = [],
    actions: list[bytes] | bytes = [],
) -> None:
    global trglist
    trglist.append(tt.Trigger(players, conditions, actions))


def copy_deaths(
    iplayer: tt.Player,
    oplayer: tt.Player,
    copyepd: bool = False,
    initvalue: int | None = None,
) -> None:
    if initvalue is None:
        if copyepd:
            initvalue = tt.EPD(0)
        else:
            initvalue = 0

    _Trigger(actions=tt.SetDeaths(oplayer, tt.SetTo, initvalue, 0))

    for i in ut._rand_lst(range(2, 32)):
        if copyepd:
            subval = 2 ** (i - 2)

        else:
            subval = 2**i

        _Trigger(
            conditions=[tt.DeathsX(iplayer, tt.AtLeast, 1, 0, 2**i)],
            actions=tt.SetDeaths(oplayer, tt.Add, subval, 0),
        )


def create_vector_relocator(chkt: CHK, payload: Payload) -> None:
    global trglist

    strmap = get_string_map()
    if strmap is None:
        raise ut.EPError(_("Must use LoadMap first"))
    str_section = strmap.save_tbl()

    """
    Algorithm credit to klassical_31@naver.com

    Overall algorithm : STR <-> MRGN cycle

    MRGN:
      SetMemory(payload, Add, payload_offset // 4)

    MRGN <-> STR0  =  Trigger(actions=[
             STR1        [SetMemory(
                            mrgn.action[i].player,
                            Add,
                            prttable[j+packn] - prttable[j]
                          ) for i in range(packn)],
             STR2        SetNextPtr(mrgn, STR[k+1])
             ...      )
    """

    mrgn = 0x58DC60
    packn = 15

    #############
    # STR SECTION
    #############
    str_sled: list[bytes] = []
    mrgn_ort = mrgn + 836
    sledheader_prt = b"\0\0\0\0" + ut.i2b4(mrgn)
    sledheader_ort = b"\0\0\0\0" + ut.i2b4(mrgn_ort)

    # apply prt
    prttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.prttable), packn):
        prts = list(map(lambda x: x // 4, payload.prttable[i : i + packn]))
        prts = prts + [-1] * (packn - len(prts))  # -1 : dummy space
        pch = [prts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(
            sledheader_prt
            + tt.Trigger(
                players=[tt.AllPlayers],
                actions=[
                    tt.SetMemory(mrgn + 328 + 32 * j + 16, tt.Add, pch[j])
                    for j in range(packn)
                ],
            )
        )

        prevoffset = prts

    # apply ort
    orttrg_start = 2408 * len(str_sled)  # = 0
    # mrgn.action[i].player = EPD(payload_offset) + prevoffset[i]
    prevoffset = [-1] * packn
    for i in range(0, len(payload.orttable), packn):
        orts = list(map(lambda x: x // 4, payload.orttable[i : i + packn]))
        orts = orts + [-1] * (packn - len(orts))  # -1 : dummy space
        pch = [orts[j] - prevoffset[j] for j in range(packn)]
        str_sled.append(
            sledheader_ort
            + tt.Trigger(
                players=[tt.AllPlayers],
                actions=[
                    tt.SetMemory(mrgn_ort + 328 + 32 * j + 16, tt.Add, pch[j])
                    for j in range(packn)
                ],
            )
        )

        prevoffset = orts

    # jump to ort
    str_sled.append(
        sledheader_ort
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                tt.SetMemory(
                    mrgn_ort + 328 + 32 * j + 16,
                    tt.Add,
                    0xFFFFFFFF - prevoffset[j],
                )
                for j in range(packn)
            ]
            + [tt.SetMemory(mrgn_ort + 4, tt.Add, 4)],  # skip garbage area
        )
    )

    sled_completed = b"".join(str_sled)

    str_padding_length = -len(str_section) & 3
    strsled_offset = len(str_section) + str_padding_length + 0x191943C8
    payload_offset = strsled_offset + len(sled_completed) + 4
    str_section = (
        str_section
        + bytes(str_padding_length)
        + sled_completed
        + bytes(4)
        + payload.data
    )
    chkt.setsection(get_string_section_name(), str_section)

    ##############
    # MRGN SECTION
    ##############
    mrgn_trigger = bytearray()
    mrgn_trigger_prt = (
        bytes(4)
        + ut.i2b4(prttrg_start + strsled_offset)
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                # SetDeaths actions in MRGN initially points to EPD(payload-4)
                tt.SetMemory(payload_offset - 4, tt.Add, payload_offset // 4)
                for _ in range(packn)
            ]
            + [tt.SetMemory(mrgn + 4, tt.Add, 2408)],
        )
        + bytes(836)
    )
    mrgn_trigger_ort = (
        bytes(836 + 4)
        + ut.i2b4(orttrg_start + strsled_offset)
        + tt.Trigger(
            players=[tt.AllPlayers],
            actions=[
                tt.SetMemory(payload_offset - 4, tt.Add, payload_offset)
                for _ in range(packn)
            ]
            + [tt.SetMemory(mrgn_ort + 4, tt.Add, 2408)],
        )
    )
    for b1, b2 in zip(mrgn_trigger_prt, mrgn_trigger_ort):
        mrgn_trigger.append(b1 ^ b2)

    oldmrgnraw = chkt.getsection("MRGN")
    mrgn_section = bytes(mrgn_trigger) + oldmrgnraw[2408 + 836 :]
    if len(mrgn_section) != 5100:
        raise RuntimeError(_("MRGN section size bug"))
    chkt.setsection("MRGN", mrgn_section)

    ##############
    # TRIG SECTION
    ##############
    trglist.clear()

    # pts[player].lasttrigger->next = value(strs) + strsled_offset

    pts = 0x51A280
    curpl = 0x6509B0

    for player in ut._rand_lst(range(8)):
        triggerend = ~(0x51A284 + player * 12)

        _Trigger(
            players=[player],
            actions=[
                tt.SetMemory(curpl, tt.SetTo, ut.EPD(pts + 12 * player + 4)),
                tt.SetDeaths(9, tt.SetTo, triggerend, 0),  # Used in stage 2
            ],
        )

    # read pts[player].lasttrigger
    for e in ut._rand_lst(range(2, 32)):
        _Trigger(
            conditions=tt.DeathsX(tt.CurrentPlayer, tt.AtLeast, 1, 0, 2**e),
            actions=tt.SetDeaths(11, tt.Add, 2**e, 0),
        )

    # apply to curpl
    _Trigger(
        actions=[
            tt.SetDeaths(10, tt.SetTo, ut.EPD(4), 0),
            tt.SetMemory(curpl, tt.SetTo, ut.EPD(4)),
        ]
    )
    for e in ut._rand_lst(range(2, 32)):
        _Trigger(
            conditions=tt.DeathsX(11, tt.AtLeast, 1, 0, 2**e),
            actions=[
                # tt.SetDeaths(11, tt.Subtract, 2 ** e, 0),
                # used for nextptr recovery in stage 3
                tt.SetDeaths(10, tt.Add, 2 ** (e - 2), 0),
                tt.SetMemory(curpl, tt.Add, 2 ** (e - 2)),
            ],
        )

    # now curpl = EPD(value(ptsprev) + 4)
    # value(EPD(value(ptsprev) + 4)) = strs + payload_offset
    # copy_deaths(tt.EPD(strs), tt.CurrentPlayer, False, strsled_offset)
    _Trigger(
        actions=[
            tt.SetDeaths(tt.CurrentPlayer, tt.SetTo, strsled_offset, 0),
            tt.SetDeaths(11, tt.SetTo, 0, 0),
        ]
    )

    # Done!
    trigdata = b"".join(trglist)

    # Stage 1 created

    # -------

    # Previous rawtrigger datas

    oldtrigraw = chkt.getsection("TRIG")
    oldtrigs = [oldtrigraw[i : i + 2400] for i in range(0, len(oldtrigraw), 2400)]
    proc_trigs = []

    # Collect only enabled triggers
    for trig in oldtrigs:
        trig = bytearray(trig)
        flag = ut.b2i4(trig, 320 + 2048)
        if flag & 8:  # Trigger already disabled
            pass

        flag ^= 8  # Disable it temporarily. It will be re-enabled at stage 3
        trig[320 + 2048 : 320 + 2048 + 4] = ut.i2b4(flag)
        proc_trigs.append(bytes(trig))

    oldtrigraw_processed = b"".join(proc_trigs)
    chkt.setsection("TRIG", trigdata + oldtrigraw_processed)



from helper import *

N = 2000
arr = EUDArray(N)


@EUDFunc
def f_bubblesort():
    for i in EUDLoopRange(N):
        for j in EUDLoopRange(N - 1 - i):
            if EUDIf()(arr[i] >= arr[i + 1]):
                arr[i], arr[i + 1] = arr[i + 1], arr[i]
            EUDEndIf()

    SetCurrentPlayer(P1)


@TestInstance
def test_perfbasic():
    f_srand(0)

    for i in EUDLoopRange(N):
        arr[i] = f_dwrand()

    test_perf("Bubble sort (%s)" % N, f_bubblesort, 1)
    test_perf("Basic looping (comparison)", lambda: None, 10000000)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable
from typing import overload

from ...localize import _
from ...utils import EPError, ExprProxy, unProxy
from .. import variable as ev
from ..mapdata import GetPropertyIndex, UnitProperty
from .consttype import (
    ConstType,
    T,
    U,
    __Arg,
    __ExprProxy,
    _Arg,
    _Byte,
    _Dword,
    _ExprProxy,
)

# fmt: off


class TrgAllyStatus(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {0: "Enemy", 1: "Ally", 2: "AlliedVictory"}[self._value]
        except KeyError:
            return super().__str__()


Enemy, Ally, AlliedVictory = TrgAllyStatus(0), TrgAllyStatus(1), TrgAllyStatus(2)
AllyStatusDict: dict[ConstType, int] = {Enemy: 0, Ally: 1, AlliedVictory: 2}


class TrgComparison(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {0: "AtLeast", 1: "AtMost", 10: "Exactly"}[self._value]
        except KeyError:
            return super().__str__()


AtLeast, AtMost, Exactly = TrgComparison(0), TrgComparison(1), TrgComparison(10)
ComparisonDict: dict[ConstType, int] = {AtLeast: 0, AtMost: 1, Exactly: 10}


class TrgCount(ConstType):
    __slots__ = ()

    def __str__(self):
        if isinstance(self._value, int) and self._value == 0:
            return "All"
        return super().__str__()


All = TrgCount(0)


class TrgModifier(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {7: "SetTo", 8: "Add", 9: "Subtract"}[self._value]
        except KeyError:
            return super().__str__()



SetTo, Add, Subtract = TrgModifier(7), TrgModifier(8), TrgModifier(9)
ModifierDict: dict[ConstType, int] = {SetTo: 7, Add: 8, Subtract: 9}


class TrgOrder(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {0: "Move", 1: "Patrol", 2: "Attack"}[self._value]
        except KeyError:
            return super().__str__()


Move, Patrol, Attack = TrgOrder(0), TrgOrder(1), TrgOrder(2)
OrderDict: dict[ConstType, int] = {Move: 0, Patrol: 1, Attack: 2}


class _Player(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {
                0: "Player1", 1: "Player2", 2: "Player3", 3: "Player4",
                4: "Player5", 5: "Player6", 6: "Player7", 7: "Player8",
                8: "Player9", 9: "Player10", 10: "Player11", 11: "Player12",
                13: "CurrentPlayer",
                14: "Foes", 15: "Allies", 16: "NeutralPlayers",
                17: "AllPlayers",
                18: "Force1", 19: "Force2", 20: "Force3", 21: "Force4",
                26: "NonAlliedVictoryPlayers",
            }[self._value]
        except KeyError:
            return super().__str__()


PlayerDict: dict[ConstType, int] = {}  # update in offsetmap.scdata


class TrgProperty(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeProperty(s)


class TrgPropState(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {4: "Enable", 5: "Disable", 6: "Toggle"}[self._value]
        except KeyError:
            return super().__str__()


class TrgSwitchAction(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {4: "Set", 5: "Clear", 6: "Toggle", 11: "Random"}[self._value]
        except KeyError:
            return super().__str__()


class TrgSwitchState(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {2: "Set", 3: "Cleared"}[self._value]
        except KeyError:
            return super().__str__()


class _Toggle(TrgPropState, TrgSwitchAction):
    __slots__ = ()
    def __init__(self):
        super().__init__(6)


class _Set(TrgSwitchState, TrgSwitchAction):
    __slots__ = ()
    def __init__(self):
        super().__init__(4)  # 2 or 4


Enable, Disable, Toggle = TrgPropState(4), TrgPropState(5), _Toggle()
Set, Clear, Random = _Set(), TrgSwitchAction(5), TrgSwitchAction(11)
Cleared = TrgSwitchState(3)
PropStateDict: dict[ConstType, int] = {Enable: 4, Disable: 5, Toggle: 6}
SwitchActionDict: dict[ConstType, int] = {Set: 4, Clear: 5, Toggle: 6, Random: 11}
SwitchStateDict: dict[ConstType, int] = {Set: 2, Cleared: 3}


class TrgResource(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {0: "Ore", 1: "Gas", 2: "OreAndGas"}[self._value]
        except KeyError:
            return super().__str__()


Ore, Gas, OreAndGas = TrgResource(0), TrgResource(1), TrgResource(2)
ResourceDict: dict[ConstType, int] = {Ore: 0, Gas: 1, OreAndGas: 2}


class TrgScore(ConstType):
    __slots__ = ()

    def __str__(self):
        try:
            return {
                0: "Total", 1: "Units", 2: "Buildings", 3: "UnitsAndBuildings",
                4: "Kills", 5: "Razings", 6: "KillsAndRazings", 7: "Custom",
            }[self._value]
        except KeyError:
            return super().__str__()


class _Kills(TrgScore):
    def __init__(self) -> None:
        super().__init__(4)
        self._internalf: Callable

    # TODO: player: Player, comparison: Comparison, number: Dword, unit: Unit
    # ) -> Condition
    def __call__(self, a, b, c, d):
        return self._internalf(a, b, c, d)


Total, Units, Buildings, UnitsAndBuildings = TrgScore(0), TrgScore(1), TrgScore(2), TrgScore(3)  # noqa: E501
# Name 'Kills' is used for both condition type and score type.
# To resolve conflict, we initialize Kills differently from others.
Kills, Razings, KillsAndRazings, Custom = _Kills(), TrgScore(5), TrgScore(6), TrgScore(7)  # noqa: E501
ScoreDict: dict[ConstType, int] = {
    Total: 0, Units: 1, Buildings: 2, UnitsAndBuildings: 3,
    Kills: 4, Razings: 5, KillsAndRazings: 6, Custom: 7,
}
# fmt: on


@overload
def _EncodeConst(t: type[ConstType], s: ConstType, u: str | None = None) -> int:
    ...


@overload
def _EncodeConst(t: type[ConstType], s: T, u: str | None = None) -> T:
    ...


@overload
def _EncodeConst(t: type[ConstType], s: _ExprProxy, u: str | None = None) -> _Dword:
    ...


def _EncodeConst(t: type[ConstType], s: _Arg, u: str | None = None) -> _Dword:  # noqa: N802
    while isinstance(s, ExprProxy) and not isinstance(s, ConstType):
        s = s._value
    if isinstance(s, ConstType):
        if isinstance(s, t):
            return unProxy(s)
        raise EPError(_('"{}" is not a {}').format(s, u if u else t.__name__))
    return s


@overload
def EncodeAllyStatus(s: ConstType) -> int:
    ...


@overload
def EncodeAllyStatus(s: U) -> U:
    ...


@overload
def EncodeAllyStatus(s: __ExprProxy) -> _Byte:
    ...


def EncodeAllyStatus(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Enemy, Ally, AlliedVictory] to number [0, 1, 2]."""
    try:
        return AllyStatusDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgAllyStatus, s)  # type: ignore[return-value]


@overload
def EncodeComparison(s: ConstType) -> int:
    ...


@overload
def EncodeComparison(s: U) -> U:
    ...


@overload
def EncodeComparison(s: __ExprProxy) -> _Byte:
    ...


def EncodeComparison(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [AtLeast, AtMost, Exactly] to number [0, 1, 10]."""
    try:
        return ComparisonDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgComparison, s)  # type: ignore[return-value]


@overload
def EncodeModifier(s: ConstType) -> int:
    ...


@overload
def EncodeModifier(s: U) -> U:
    ...


@overload
def EncodeModifier(s: __ExprProxy) -> _Byte:
    ...


def EncodeModifier(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [SetTo, Add, Subtract] to number [7, 8, 9]."""
    try:
        return ModifierDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgModifier, s)  # type: ignore[return-value]


@overload
def EncodeOrder(s: ConstType) -> int:
    ...


@overload
def EncodeOrder(s: U) -> U:
    ...


@overload
def EncodeOrder(s: __ExprProxy) -> _Byte:
    ...


def EncodeOrder(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Move, Patrol, Attack] to number [0, 1, 2]."""
    try:
        return OrderDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgOrder, s)  # type: ignore[return-value]


@overload
def EncodePlayer(s: ConstType) -> int:
    ...


@overload
def EncodePlayer(s: T) -> T:
    ...


@overload
def EncodePlayer(s: _ExprProxy) -> _Dword:
    ...


def EncodePlayer(s: _Arg) -> _Dword:  # noqa: N802
    """Convert player identifier to corresponding number.

    ======================= ========
        Identifier           Number
    ======================= ========
    P1                         0
    P2                         1
    P3                         2
    P4                         3
    P5                         4
    P6                         5
    P7                         6
    P8                         7
    P9                         8
    P10                        9
    P11                        10
    P12                        11
    Player1                    0
    Player2                    1
    Player3                    2
    Player4                    3
    Player5                    4
    Player6                    5
    Player7                    6
    Player8                    7
    Player9                    8
    Player10                   9
    Player11                   10
    Player12                   11
    CurrentPlayer              13
    Foes                       14
    Allies                     15
    NeutralPlayers             16
    AllPlayers                 17
    Force1                     18
    Force2                     19
    Force3                     20
    Force4                     21
    NonAlliedVictoryPlayers    26
    ======================= ========

    """
    if type(s) is ev.EUDVariable:
        return s

    try:
        return PlayerDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(_Player, s, "TrgPlayer")


@overload
def EncodePropState(s: ConstType) -> int:
    ...


@overload
def EncodePropState(s: U) -> U:
    ...


@overload
def EncodePropState(s: __ExprProxy) -> _Byte:
    ...


def EncodePropState(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Enable, Disable, Toogle] to number [4, 5, 6]"""
    try:
        return PropStateDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgPropState, s)  # type: ignore[return-value]


@overload
def EncodeResource(s: ConstType) -> int:
    ...


@overload
def EncodeResource(s: U) -> U:
    ...


@overload
def EncodeResource(s: __ExprProxy) -> _Byte:
    ...


def EncodeResource(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Ore, Gas, OreAndGas] to [0, 1, 2]"""
    try:
        return ResourceDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgResource, s)  # type: ignore[return-value]


@overload
def EncodeScore(s: ConstType) -> int:
    ...


@overload
def EncodeScore(s: U) -> U:
    ...


@overload
def EncodeScore(s: __ExprProxy) -> _Byte:
    ...


def EncodeScore(s: __Arg) -> _Byte:  # noqa: N802
    """Convert score type identifier to number.

    ================= ========
        _Score type     Number
    ================= ========
    Total                0
    Units                1
    Buildings            2
    UnitsAndBuildings    3
    Kills                4
    Razings              5
    KillsAndRazings      6
    Custom               7
    ================= ========

    """
    try:
        return ScoreDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgScore, s)  # type: ignore[return-value]


@overload
def EncodeSwitchAction(s: ConstType) -> int:
    ...


@overload
def EncodeSwitchAction(s: U) -> U:
    ...


@overload
def EncodeSwitchAction(s: __ExprProxy) -> _Byte:
    ...


def EncodeSwitchAction(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Set, Clear, Toogle, Random] to [4, 5, 6, 11]."""
    try:
        return SwitchActionDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgSwitchAction, s)  # type: ignore[return-value]


@overload
def EncodeSwitchState(s: ConstType) -> int:
    ...


@overload
def EncodeSwitchState(s: U) -> U:
    ...


@overload
def EncodeSwitchState(s: __ExprProxy) -> _Byte:
    ...


def EncodeSwitchState(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [Set, Cleared] to [2, 3]."""
    try:
        return SwitchStateDict[s]  # type: ignore[index]
    except KeyError:
        return _EncodeConst(TrgSwitchState, s)  # type: ignore[return-value]


@overload
def EncodeCount(s: ConstType) -> int:
    ...


@overload
def EncodeCount(s: U) -> U:
    ...


@overload
def EncodeCount(s: __ExprProxy) -> _Byte:
    ...


def EncodeCount(s: __Arg) -> _Byte:  # noqa: N802
    """Convert [All, (other numbers)] to number [0, (as-is)]."""
    if s is All:
        return 0
    return _EncodeConst(TrgCount, s)  # type: ignore[return-value]


# ========================


def EncodeProperty(  # noqa: N802
    prop: UnitProperty | bytes | ExprProxy[UnitProperty | bytes],
) -> int:
    prop = unProxy(prop)
    return GetPropertyIndex(prop)



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools

from .. import core as c
from ..core import Add, Memory, SetMemory
from ..core.eudfunc.eudf import _EUDPredefineParam
from ..core.eudstruct.vararray import _InternalVArray
from ..utils import EPD, EUDCreateBlock, EUDPopBlock, ep_assert


@functools.cache
def EUDQueue(capacity):  # noqa: N802
    """A single-ended queue implemented with a fixed-size variable array."""
    ep_assert(
        isinstance(capacity, int) and capacity > 0,
        "capacity of EUDQueue should be greater than 0",
    )

    class _EUDQueue:
        def __init__(self):  # TODO: support initval
            ret = c.EUDVariable()
            pop = c.EUDXVariable(ret, c.SetTo, 0)
            queue = _InternalVArray(capacity)(
                dest=EPD(pop.getValueAddr()), nextptr=pop.GetVTable()
            )
            append_act = SetMemory(queue + 348, c.SetTo, 0)
            jumpleft = c.Forward()
            self._length = c.EUDVariable(0)

            c.PushTriggerScope()
            iter_start, iter_jump = c.Forward(), c.Forward()
            iter_init = c.RawTrigger(
                actions=[
                    SetMemory(iter_start + 16, c.SetTo, EPD(queue) + 87),
                    c.SetNextPtr(iter_jump, queue - 72),
                    c.SetNextPtr(iter_start, iter_jump),
                    c.SetMemoryX(iter_start + 2376, c.SetTo, 1, 1),
                ],
            )
            c.RawTrigger(
                conditions=self._length.AtLeast(1),
                actions=c.SetMemoryX(iter_start + 2404, c.SetTo, 1 << 24, 255 << 24),
            )
            iter_start << c.RawTrigger(
                nextptr=0,  # by iter_init
                conditions=Memory(
                    append_act + 16, c.Exactly, 0
                ),  # by iter_init, iter_contpoint
                actions=c.SetNextPtr(iter_start, 0),  # end by __iter__ call
            )
            iter_jump << c.RawTrigger(
                nextptr=0,  # by iter_init, iter_contpoint
                actions=[
                    c.SetNextPtr(pop.GetVTable(), 0),  # yield_point by __iter__ call
                    SetMemory(iter_start + 16, Add, 18),
                    SetMemory(iter_jump + 4, Add, 72),
                    c.SetMemoryX(iter_start + 2376, c.SetTo, 0, 1),
                ],
            )
            iter_contpoint = c.RawTrigger(
                nextptr=iter_start,
                conditions=Memory(
                    iter_start + 16, c.AtLeast, EPD(queue) + 87 + 18 * capacity
                ),
                actions=[
                    SetMemory(iter_start + 16, Add, -(18 * capacity)),
                    SetMemory(iter_jump + 4, Add, -(72 * capacity)),
                ],
            )
            c.PopTriggerScope()

            def iter():
                yield_point, end = c.Forward(), c.Forward()
                block = {
                    "loopstart": iter_start,
                    "loopend": end,
                    "contpoint": c.Forward(),
                }
                EUDCreateBlock("loopqueueblock", block)
                c.RawTrigger(
                    nextptr=iter_init,
                    actions=[
                        SetMemory(iter_start + 348, c.SetTo, end),
                        SetMemory(iter_jump + 348, c.SetTo, yield_point),
                    ],
                )

                yield_point << c.NextTrigger()
                yield ret

                block = EUDPopBlock("loopqueueblock")[1]
                if not block["contpoint"].IsSet():
                    block["contpoint"] << iter_contpoint
                c.SetNextTrigger(iter_contpoint)
                end << c.NextTrigger()

            @_EUDPredefineParam((EPD(append_act) + 5,))
            @c.EUDFunc
            def append(value):
                check_wrap, wrap_tail, wrap_head = (
                    c.Forward(),
                    c.Forward(),
                    c.Forward(),
                )
                c.RawTrigger(
                    actions=[
                        append_act,
                        SetMemory(append_act + 16, Add, 18),
                        self._length.AddNumber(1),
                    ],
                )
                check_wrap << c.RawTrigger(
                    nextptr=wrap_head,
                    conditions=self._length.AtLeast(capacity + 1),
                    actions=[
                        c.SetNextPtr(check_wrap, wrap_tail),
                        SetMemory(jumpleft + 4, Add, 72),
                        SetMemory(iter_init + 348, Add, 18),
                        SetMemory(iter_init + 380, Add, 72),
                        self._length.SubtractNumber(1),
                    ],
                )
                wrap_tail << c.RawTrigger(
                    conditions=Memory(
                        jumpleft + 4, c.AtLeast, queue + 72 * (capacity - 1)
                    ),
                    actions=[
                        c.SetNextPtr(check_wrap, wrap_head),
                        SetMemory(jumpleft + 4, Add, -(72 * capacity)),
                        SetMemory(iter_init + 348, Add, -(18 * capacity)),
                        SetMemory(iter_init + 380, Add, -(72 * capacity)),
                    ],
                )
                wrap_head << c.RawTrigger(
                    conditions=Memory(
                        append_act + 16,
                        c.AtLeast,
                        EPD(queue) + 87 + 18 * capacity,
                    ),
                    actions=SetMemory(append_act + 16, Add, -(18 * capacity)),
                )

            @c.EUDFunc
            def popleft():
                # ret = queue[tail]
                # tail %+ 1
                # return ret
                nonlocal jumpleft
                wraparound = c.Forward()
                jumpleft << c.RawTrigger(
                    nextptr=queue - 72,
                    actions=[
                        c.SetNextPtr(pop.GetVTable(), wraparound),
                        SetMemory(jumpleft + 4, Add, 72),
                        SetMemory(iter_init + 348, Add, 18),
                        SetMemory(iter_init + 380, Add, 72),
                        self._length.SubtractNumber(1),
                    ],
                )
                wraparound << c.RawTrigger(
                    conditions=Memory(
                        jumpleft + 4, c.AtLeast, queue + 72 * (capacity - 1)
                    ),
                    actions=[
                        SetMemory(jumpleft + 4, Add, -(72 * capacity)),
                        SetMemory(iter_init + 348, Add, -(18 * capacity)),
                        SetMemory(iter_init + 380, Add, -(72 * capacity)),
                    ],
                )
                return ret

            def clear():
                c.RawTrigger(
                    actions=[
                        SetMemory(jumpleft + 4, c.SetTo, queue - 72),
                        SetMemory(iter_init + 348, c.SetTo, EPD(queue) + 87),
                        SetMemory(iter_init + 380, c.SetTo, queue - 72),
                        SetMemory(append_act + 16, c.SetTo, EPD(queue) + 87),
                        self._length.SetNumber(0),
                    ]
                )

            self._append = append
            self._popleft = popleft
            self._iter = iter
            self._clear = clear

            c.PushTriggerScope()
            append(0)  # prevent 'Cannot evaluate <ConstExpr>...' error
            popleft()  # by always instantiate triggers
            c.PopTriggerScope()

        def __iter__(self):
            return self._iter()

        def append(self, value, **kwargs):
            self._append(value, **kwargs)

        def popleft(self, **kwargs):
            return self._popleft(**kwargs)

        def clear(self):
            self._clear()

        def empty(self):
            return self._length == 0

        @property  # FIXME: eudplib code can mutate length
        def length(self):
            return self._length

        def eqattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length == k
            raise AttributeError

        def neattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length != k
            raise AttributeError

        def leattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length <= k
            raise AttributeError

        def ltattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length < k
            raise AttributeError

        def geattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length >= k
            raise AttributeError

        def gtattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length > k
            raise AttributeError

    return _EUDQueue


@functools.cache
def EUDDeque(capacity):  # noqa: N802
    """A double-ended queue implemented with a fixed-size variable array."""
    ep_assert(
        isinstance(capacity, int) and capacity > 0,
        "capacity of EUDDeque should be greater than 0",
    )

    class _EUDDeque:
        def __init__(self):  # TODO: support initval
            ret = c.EUDVariable()
            pop = c.EUDXVariable(ret, c.SetTo, 0)
            deque = _InternalVArray(capacity)(
                dest=EPD(pop.getValueAddr()), nextptr=pop.GetVTable()
            )
            append_act = SetMemory(deque + 348, c.SetTo, 0)
            appendleft_act = SetMemory(deque + 348, c.SetTo, 0)
            jump, jumpleft = c.Forward(), c.Forward()
            self._length = c.EUDVariable(0)

            c.PushTriggerScope()
            iter_start, iter_jump = c.Forward(), c.Forward()
            iter_init = c.RawTrigger(
                actions=[
                    SetMemory(iter_start + 16, c.SetTo, EPD(deque) + 87),
                    c.SetNextPtr(iter_jump, deque - 72),
                    c.SetNextPtr(iter_start, iter_jump),
                    c.SetMemoryX(iter_start + 2376, c.SetTo, 1, 1),
                ],
            )
            c.RawTrigger(
                conditions=self._length.AtLeast(1),
                actions=c.SetMemoryX(iter_start + 2404, c.SetTo, 1 << 24, 255 << 24),
            )
            iter_start << c.RawTrigger(
                nextptr=0,  # by iter_init
                conditions=Memory(
                    append_act + 16, c.Exactly, 0
                ),  # by iter_init, iter_contpoint
                actions=c.SetNextPtr(iter_start, 0),  # end by __iter__ call
            )
            iter_jump << c.RawTrigger(
                nextptr=0,  # by iter_init, iter_contpoint
                actions=[
                    c.SetNextPtr(pop.GetVTable(), 0),  # yield_point by __iter__ call
                    SetMemory(iter_start + 16, Add, 18),
                    SetMemory(iter_jump + 4, Add, 72),
                    c.SetMemoryX(iter_start + 2376, c.SetTo, 0, 1),
                ],
            )
            iter_contpoint = c.RawTrigger(
                nextptr=iter_start,
                conditions=Memory(
                    iter_start + 16, c.AtLeast, EPD(deque) + 87 + 18 * capacity
                ),
                actions=[
                    SetMemory(iter_start + 16, Add, -(18 * capacity)),
                    SetMemory(iter_jump + 4, Add, -(72 * capacity)),
                ],
            )
            c.PopTriggerScope()

            def iter():
                yield_point, end = c.Forward(), c.Forward()
                block = {
                    "loopstart": iter_start,
                    "loopend": end,
                    "contpoint": c.Forward(),
                }
                EUDCreateBlock("loopqueueblock", block)
                c.RawTrigger(
                    nextptr=iter_init,
                    actions=[
                        SetMemory(iter_start + 348, c.SetTo, end),
                        SetMemory(iter_jump + 348, c.SetTo, yield_point),
                    ],
                )

                yield_point << c.NextTrigger()
                yield ret

                block = EUDPopBlock("loopqueueblock")[1]
                if not block["contpoint"].IsSet():
                    block["contpoint"] << iter_contpoint
                c.SetNextTrigger(iter_contpoint)
                end << c.NextTrigger()

            @_EUDPredefineParam((EPD(append_act) + 5,))
            @c.EUDFunc
            def append(value):
                # deque[head] = value
                # head %+ 1
                check_wrap, wrap_tail, wrap_head = (
                    c.Forward(),
                    c.Forward(),
                    c.Forward(),
                )
                c.RawTrigger(
                    actions=[
                        append_act,
                        SetMemory(append_act + 16, Add, 18),
                        SetMemory(jump + 4, Add, 72),
                        self._length.AddNumber(1),
                        c.SetNextPtr(check_wrap, wrap_head),
                    ],
                )
                check_wrap << c.RawTrigger(
                    nextptr=wrap_head,
                    conditions=self._length.AtLeast(capacity + 1),
                    actions=[
                        c.SetNextPtr(check_wrap, wrap_tail),
                        SetMemory(appendleft_act + 16, Add, 18),
                        SetMemory(jumpleft + 4, Add, 72),
                        SetMemory(iter_init + 348, Add, 18),
                        SetMemory(iter_init + 380, Add, 72),
                        self._length.SubtractNumber(1),
                    ],
                )
                wrap_tail << c.RawTrigger(
                    conditions=Memory(
                        jumpleft + 4, c.AtLeast, deque + 72 * (capacity - 1)
                    ),
                    actions=[
                        SetMemory(appendleft_act + 16, Add, -(18 * capacity)),
                        SetMemory(jumpleft + 4, Add, -(72 * capacity)),
                        SetMemory(iter_init + 348, Add, -(18 * capacity)),
                        SetMemory(iter_init + 380, Add, -(72 * capacity)),
                    ],
                )
                wrap_head << c.RawTrigger(
                    conditions=Memory(
                        append_act + 16,
                        c.AtLeast,
                        EPD(deque) + 87 + 18 * capacity,
                    ),
                    actions=[
                        SetMemory(append_act + 16, Add, -(18 * capacity)),
                        SetMemory(jump + 4, Add, -(72 * capacity)),
                    ],
                )

            @c.EUDFunc
            def _pop():
                # head %- 1
                # return deque[head]
                nonlocal jump
                retpoint = c.Forward()
                c.RawTrigger(
                    conditions=Memory(jump + 4, c.AtMost, deque),
                    actions=[
                        SetMemory(append_act + 16, Add, 18 * capacity),
                        SetMemory(jump + 4, Add, 72 * capacity),
                    ],
                )
                jump << c.RawTrigger(
                    nextptr=deque,
                    actions=[
                        c.SetNextPtr(pop.GetVTable(), retpoint),
                        SetMemory(append_act + 16, Add, -18),
                        SetMemory(jump + 4, Add, -72),
                        self._length.SubtractNumber(1),
                    ],
                )
                retpoint << c.NextTrigger()
                return ret

            @_EUDPredefineParam((EPD(appendleft_act) + 5,))
            @c.EUDFunc
            def appendleft(value):
                # tail %- 1
                # deque[tail] = value
                check_wrap, wrap_head, wrap_tail = (
                    c.Forward(),
                    c.Forward(),
                    c.Forward(),
                )
                check_wrap << c.RawTrigger(
                    nextptr=wrap_tail,
                    conditions=self._length.AtLeast(capacity),
                    actions=[
                        c.SetNextPtr(check_wrap, wrap_head),
                        SetMemory(append_act + 16, Add, -18),
                        SetMemory(jump + 4, Add, -72),
                        self._length.SubtractNumber(1),
                    ],
                )
                wrap_head << c.RawTrigger(
                    conditions=Memory(jump + 4, c.AtMost, deque),
                    actions=[
                        SetMemory(append_act + 16, Add, 18 * capacity),
                        SetMemory(jump + 4, Add, 72 * capacity),
                    ],
                )
                wrap_tail << c.RawTrigger(
                    conditions=Memory(
                        appendleft_act + 16, c.AtMost, EPD(deque) + 87
                    ),
                    actions=[
                        SetMemory(appendleft_act + 16, Add, 18 * capacity),
                        SetMemory(jumpleft + 4, Add, 72 * capacity),
                        SetMemory(iter_init + 348, Add, 18 * capacity),
                        SetMemory(iter_init + 380, Add, 72 * capacity),
                    ],
                )
                c.RawTrigger(
                    actions=[
                        c.SetNextPtr(check_wrap, wrap_tail),
                        SetMemory(appendleft_act + 16, Add, -18),
                        SetMemory(jumpleft + 4, Add, -72),
                        SetMemory(iter_init + 348, Add, -18),
                        SetMemory(iter_init + 380, Add, -72),
                        appendleft_act,
                        self._length.AddNumber(1),
                    ],
                )

            @c.EUDFunc
            def popleft():
                # ret = deque[tail]
                # tail %+ 1
                # return ret
                nonlocal jumpleft
                wraparound = c.Forward()
                jumpleft << c.RawTrigger(
                    nextptr=deque - 72,
                    actions=[
                        c.SetNextPtr(pop.GetVTable(), wraparound),
                        SetMemory(appendleft_act + 16, Add, 18),
                        SetMemory(jumpleft + 4, Add, 72),
                        SetMemory(iter_init + 348, Add, 18),
                        SetMemory(iter_init + 380, Add, 72),
                        self._length.SubtractNumber(1),
                    ],
                )
                wraparound << c.RawTrigger(
                    conditions=Memory(
                        jumpleft + 4, c.AtLeast, deque + 72 * (capacity - 1)
                    ),
                    actions=[
                        SetMemory(appendleft_act + 16, Add, -(18 * capacity)),
                        SetMemory(jumpleft + 4, Add, -(72 * capacity)),
                        SetMemory(iter_init + 348, Add, -(18 * capacity)),
                        SetMemory(iter_init + 380, Add, -(72 * capacity)),
                    ],
                )
                return ret

            def clear():
                c.RawTrigger(
                    actions=[
                        SetMemory(jump + 4, c.SetTo, deque),
                        SetMemory(jumpleft + 4, c.SetTo, deque - 72),
                        SetMemory(iter_init + 348, c.SetTo, EPD(deque) + 87),
                        SetMemory(iter_init + 380, c.SetTo, deque - 72),
                        SetMemory(append_act + 16, c.SetTo, EPD(deque) + 87),
                        SetMemory(appendleft_act + 16, c.SetTo, EPD(deque) + 87),
                        self._length.SetNumber(0),
                    ]
                )

            self._append = append
            self._appendleft = appendleft
            self._pop = _pop
            self._popleft = popleft
            self._iter = iter
            self._clear = clear

            c.PushTriggerScope()
            append(0)  # prevent 'Cannot evaluate <ConstExpr>...' error
            _pop()  # by always instantiate triggers
            appendleft(0)
            popleft()
            c.PopTriggerScope()

        def __iter__(self):
            return self._iter()

        def append(self, value, **kwargs):
            self._append(value, **kwargs)

        def pop(self, **kwargs):
            return self._pop(**kwargs)

        def appendleft(self, value, **kwargs):
            self._appendleft(value, **kwargs)

        def popleft(self, **kwargs):
            return self._popleft(**kwargs)

        def clear(self):
            self._clear()

        def empty(self):
            return self._length == 0

        @property  # FIXME: eudplib code can mutate length
        def length(self):
            return self._length

        def eqattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length == k
            raise AttributeError

        def neattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length != k
            raise AttributeError

        def leattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length <= k
            raise AttributeError

        def ltattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length < k
            raise AttributeError

        def geattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length >= k
            raise AttributeError

        def gtattr(self, attr, k):
            if isinstance(attr, str) and attr == "length":
                return self._length > k
            raise AttributeError

    return _EUDDeque



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .mapdata import GetChkTokenized, IsMapdataInitialized
from .playerinfo import GetPlayerInfo
from .proptable import GetPropertyIndex
from .stringmap import GetLocationIndex, GetStringIndex, GetSwitchIndex, GetUnitIndex
from .tblformat import TBL
from .unitprp import UnitProperty

__all__ = [
    "GetChkTokenized",
    "IsMapdataInitialized",
    "GetPlayerInfo",
    "GetPropertyIndex",
    "GetLocationIndex",
    "GetStringIndex",
    "GetSwitchIndex",
    "GetUnitIndex",
    "TBL",
    "UnitProperty",
]
